from . import _iriplas
import numpy as np
from pkg_resources import resource_filename
from scipy.interpolate import interp1d


class pyiriplas:
    def __init__(self,year=2016,month=5,day=12,hour=1.,lon=0.,lat=0.):
        self.lon=lon
        self.lat=lat
        self.yr=year
        self.month=month
        self.day=day
        self.hour=hour
        self.jmag=0 #geographic (0) or geomagnetic (1) coordinates
        self.datapath=resource_filename(__name__,'ig_rz.dat').replace('ig_rz.dat','')
        

    def get_profile(self,heights):
        '''get interpolated density profile at heights (between 80 and 20200km)'''
        #res=_iri.iri_sub(self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=hstart,heiend=hend,heistp=hstep)[0][0]
        res=_iriplas.iri_plas_main(self.datapath,alati=self.lat,along=self.lon,jmag=self.jmag,iyyyy=self.yr,mmdd=self.month*100+self.day,hours=self.hour)
        h = res[0]
        hidx = h>0
        NE = res[1][hidx]
        h=h[hidx]
        hinterpol = interp1d(h,NE,fill_value='extrapolate')
        
        return hinterpol(heights)




# This file was automatically generated by SWIG (http://www.swig.org).
# Version 2.0.4
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.



from sys import version_info
if version_info >= (2,6,0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_EMM_Model', [dirname(__file__)])
        except ImportError:
            import _EMM_Model
            return _EMM_Model
        if fp is not None:
            try:
                _mod = imp.load_module('_EMM_Model', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _EMM_Model = swig_import_helper()
    del swig_import_helper
else:
    import _EMM_Model
del version_info
try:
    _swig_property = property
except NameError:
    pass # Python < 2.2 doesn't have 'property'.
def _swig_setattr_nondynamic(self,class_type,name,value,static=1):
    if (name == "thisown"): return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name,None)
    if method: return method(self,value)
    if (not static):
        self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)

def _swig_setattr(self,class_type,name,value):
    return _swig_setattr_nondynamic(self,class_type,name,value,0)

def _swig_getattr(self,class_type,name):
    if (name == "thisown"): return self.this.own()
    method = class_type.__swig_getmethods__.get(name,None)
    if method: return method(self)
    raise AttributeError(name)

def _swig_repr(self):
    try: strthis = "proxy of " + self.this.__repr__()
    except: strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object : pass
    _newclass = 0


class WMM_Model(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, WMM_Model, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, WMM_Model, name)
    __repr__ = _swig_repr
    __swig_destroy__ = _EMM_Model.delete_WMM_Model
    __del__ = lambda self : None;
    def __init__(self, *args): 
        this = _EMM_Model.new_WMM_Model(*args)
        try: self.this.append(this)
        except: self.this = this
    def setEM(self): return _EMM_Model.WMM_Model_setEM(self)
    def setDate(self, *args): return _EMM_Model.WMM_Model_setDate(self, *args)
    def setLonLat(self, *args): return _EMM_Model.WMM_Model_setLonLat(self, *args)
    def setHeight(self, *args): return _EMM_Model.WMM_Model_setHeight(self, *args)
    def getX(self): return _EMM_Model.WMM_Model_getX(self)
    def getY(self): return _EMM_Model.WMM_Model_getY(self)
    def getZ(self): return _EMM_Model.WMM_Model_getZ(self)
    def getdX(self): return _EMM_Model.WMM_Model_getdX(self)
    def getdY(self): return _EMM_Model.WMM_Model_getdY(self)
    def getdZ(self): return _EMM_Model.WMM_Model_getdZ(self)
    def getDate(self): return _EMM_Model.WMM_Model_getDate(self)
    def getLon(self): return _EMM_Model.WMM_Model_getLon(self)
    def getLat(self): return _EMM_Model.WMM_Model_getLat(self)
    def getHeight(self): return _EMM_Model.WMM_Model_getHeight(self)
    __swig_setmethods__["NumTerms"] = _EMM_Model.WMM_Model_NumTerms_set
    __swig_getmethods__["NumTerms"] = _EMM_Model.WMM_Model_NumTerms_get
    if _newclass:NumTerms = _swig_property(_EMM_Model.WMM_Model_NumTerms_get, _EMM_Model.WMM_Model_NumTerms_set)
    __swig_setmethods__["LoadedEpoch"] = _EMM_Model.WMM_Model_LoadedEpoch_set
    __swig_getmethods__["LoadedEpoch"] = _EMM_Model.WMM_Model_LoadedEpoch_get
    if _newclass:LoadedEpoch = _swig_property(_EMM_Model.WMM_Model_LoadedEpoch_get, _EMM_Model.WMM_Model_LoadedEpoch_set)
    __swig_setmethods__["nMax"] = _EMM_Model.WMM_Model_nMax_set
    __swig_getmethods__["nMax"] = _EMM_Model.WMM_Model_nMax_get
    if _newclass:nMax = _swig_property(_EMM_Model.WMM_Model_nMax_get, _EMM_Model.WMM_Model_nMax_set)
    __swig_setmethods__["nMaxEMM"] = _EMM_Model.WMM_Model_nMaxEMM_set
    __swig_getmethods__["nMaxEMM"] = _EMM_Model.WMM_Model_nMaxEMM_get
    if _newclass:nMaxEMM = _swig_property(_EMM_Model.WMM_Model_nMaxEMM_get, _EMM_Model.WMM_Model_nMaxEMM_set)
    __swig_setmethods__["Ellip"] = _EMM_Model.WMM_Model_Ellip_set
    __swig_getmethods__["Ellip"] = _EMM_Model.WMM_Model_Ellip_get
    if _newclass:Ellip = _swig_property(_EMM_Model.WMM_Model_Ellip_get, _EMM_Model.WMM_Model_Ellip_set)
    __swig_setmethods__["CoordSpherical"] = _EMM_Model.WMM_Model_CoordSpherical_set
    __swig_getmethods__["CoordSpherical"] = _EMM_Model.WMM_Model_CoordSpherical_get
    if _newclass:CoordSpherical = _swig_property(_EMM_Model.WMM_Model_CoordSpherical_get, _EMM_Model.WMM_Model_CoordSpherical_set)
    __swig_setmethods__["CoordGeodetic"] = _EMM_Model.WMM_Model_CoordGeodetic_set
    __swig_getmethods__["CoordGeodetic"] = _EMM_Model.WMM_Model_CoordGeodetic_get
    if _newclass:CoordGeodetic = _swig_property(_EMM_Model.WMM_Model_CoordGeodetic_get, _EMM_Model.WMM_Model_CoordGeodetic_set)
    __swig_setmethods__["UserDate"] = _EMM_Model.WMM_Model_UserDate_set
    __swig_getmethods__["UserDate"] = _EMM_Model.WMM_Model_UserDate_get
    if _newclass:UserDate = _swig_property(_EMM_Model.WMM_Model_UserDate_get, _EMM_Model.WMM_Model_UserDate_set)
    __swig_setmethods__["GeoMagneticElements"] = _EMM_Model.WMM_Model_GeoMagneticElements_set
    __swig_getmethods__["GeoMagneticElements"] = _EMM_Model.WMM_Model_GeoMagneticElements_get
    if _newclass:GeoMagneticElements = _swig_property(_EMM_Model.WMM_Model_GeoMagneticElements_get, _EMM_Model.WMM_Model_GeoMagneticElements_set)
    __swig_setmethods__["Geoid"] = _EMM_Model.WMM_Model_Geoid_set
    __swig_getmethods__["Geoid"] = _EMM_Model.WMM_Model_Geoid_get
    if _newclass:Geoid = _swig_property(_EMM_Model.WMM_Model_Geoid_get, _EMM_Model.WMM_Model_Geoid_set)
WMM_Model_swigregister = _EMM_Model.WMM_Model_swigregister
WMM_Model_swigregister(WMM_Model)

class EMM_Model(WMM_Model):
    __swig_setmethods__ = {}
    for _s in [WMM_Model]: __swig_setmethods__.update(getattr(_s,'__swig_setmethods__',{}))
    __setattr__ = lambda self, name, value: _swig_setattr(self, EMM_Model, name, value)
    __swig_getmethods__ = {}
    for _s in [WMM_Model]: __swig_getmethods__.update(getattr(_s,'__swig_getmethods__',{}))
    __getattr__ = lambda self, name: _swig_getattr(self, EMM_Model, name)
    __repr__ = _swig_repr
    def __init__(self, *args): 
        this = _EMM_Model.new_EMM_Model(*args)
        try: self.this.append(this)
        except: self.this = this
    def setEM(self): return _EMM_Model.EMM_Model_setEM(self)
    __swig_destroy__ = _EMM_Model.delete_EMM_Model
    __del__ = lambda self : None;
EMM_Model_swigregister = _EMM_Model.EMM_Model_swigregister
EMM_Model_swigregister(EMM_Model)

# This file is compatible with both classic and new-style classes.





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Extract TEC values from an IONEX file given a specific time and geographic coordinate.

Created on Tue Apr 24 11:46:57 2018

@author: mevius
"""
import numpy as np
import datetime
import scipy.ndimage.filters as myfilter
import logging
import os
import ftplib
import socket


logging.basicConfig(level=logging.ERROR)


def _read_ionex_header(filep):
    """reads header from ionex file. returns data shape and position of first
    data in the file.

    Args:
        filep (filepointer) : pointer to opened ionex file.

    Returns:
        Tuple[float, np.array, np.array, np.array]:
            multiplication factor,lonarray,latarray,timearray
    """

    filep.seek(0)
    for line in filep:
        if "END OF HEADER" in line:
            break
        stripped = line.strip()
        if stripped.endswith("EPOCH OF FIRST MAP"):
            starttime = datetime.datetime(
                *(int(i) for i in
                  stripped.replace("EPOCH OF FIRST MAP","").split()))
        if stripped.endswith("EPOCH OF LAST MAP"):
            endtime = datetime.datetime(
                *(int(i) for i in
                  stripped.replace("EPOCH OF LAST MAP","").split()))
        if stripped.endswith("INTERVAL"):
            timestep = float(stripped.split()[0]) / 3600.
        if stripped.endswith("EXPONENT"):
            exponent = pow(10, float(stripped.split()[0]))
        if stripped.endswith("DLON"):
            start_lon, end_lon, step_lon = \
                (float(i) for i in stripped.split()[:3])
        if stripped.endswith("DLAT"):
            start_lat, end_lat, step_lat = \
                (float(i) for i in stripped.split()[:3])
        if stripped.endswith("OF MAPS IN FILE"):
            ntimes = int(stripped.split()[0])

    lonarray = np.arange(start_lon, end_lon + step_lon, step_lon)
    latarray = np.arange(start_lat, end_lat + step_lat, step_lat)
    dtime = endtime - starttime
    dtimef = dtime.days * 24. + dtime.seconds / 3600.
    logging.debug("timerange %f hours. step = %f ", dtimef, timestep)
    timearray = np.arange(0,
                          dtimef + timestep,
                          timestep)
    if timearray.shape[0] < ntimes:
        # bug in ILTF files,last time in header is incorrect
        extratimes = np.arange(timearray[-1] + timestep,
                               timearray[-1]
                               + (ntimes -
                                  timearray.shape[0] + 0.5) * timestep,
                               timestep)
        timearray = np.concatenate((timearray, extratimes))
    timearray += starttime.hour\
        + starttime.minute/60.\
        + starttime.second/3600.

    return exponent, lonarray, latarray, timearray


def read_tec(filename, _use_filter=None):
    """ returns TEC, RMS longitude, lattitude and time read from an IONEX file.

    Args:
        filename (string) : the full path to the IONEXfile
        _use_filter (float) : optional filter the data in space and time
                             with a gaussian filter with sigma use_filter.
                             calls scipy.ndimage.filter.gaussian_filter(tec,
                             use_filter)

    Returns:
        Tuple[np.array, np.array, np.array, np.array, np.array]:
            3D-arrays (time,lat,lon) of (optionally filtered) TEC and RMS +
            longitude, latitude and time array
    """
    ionex_file = open(filename, "r")
    exponent, lonarray, latarray, timearray = _read_ionex_header(ionex_file)
    logging.info("reading data with shapes %d  x %d x %d",
                 timearray.shape[0],
                 latarray.shape[0],
                 lonarray.shape[0])
    tecarray = np.zeros(timearray.shape
                        + latarray.shape + lonarray.shape, dtype=float)
    rmsarray = np.zeros_like(tecarray)
    timeidx = 0
    lonidx = 0
    latidx = 0
    tecdata = False
    rmsdata = False
    readdata = False
    for line in ionex_file:
        if "START OF TEC MAP" in line:
            tecdata = True
            rmsdata = False
            timeidx = int(line.strip().split()[0]) - 1
            continue
        if "START OF RMS MAP" in line:
            rmsdata = True
            tecdata = False
            timeidx = int(line.strip().split()[0]) - 1
            continue
        if "LAT/LON1/LON2/DLON/H" in line:
            readdata = True
            latstr = line.strip().replace("LAT/LON1/LON2/DLON/H","")
            lat = np.fromstring(" -".join(latstr.split("-")), sep=" ")
            latidx = np.argmin(np.abs(latarray - lat[0]))
            lonidx = 0
            continue
        if tecdata and ("END OF TEC MAP" in line):
            readdata = False
            continue
        if rmsdata and ("END OF RMS MAP" in line):
            readdata = False
            continue
        if readdata:
            data = np.fromstring(" -".join(line.strip().split("-")),
                                 sep=" ") * exponent
            if tecdata:
                tecarray[timeidx, latidx, lonidx:lonidx + data.shape[0]] = data
            elif rmsdata:
                rmsarray[timeidx, latidx, lonidx:lonidx + data.shape[0]] = data
            lonidx += data.shape[0]
    if not _use_filter is None:
        tecarray = myfilter.gaussian_filter(
            tecarray, _use_filter, mode='nearest')
    return tecarray, rmsarray, lonarray, latarray, timearray


def readTEC(filename, use_filter=None):
    """oldfunction name for compatibility. Use read_tec."""

    logging.warning("function readTEC obsolete, use read_tec instead")
    return read_tec(filename, _use_filter=use_filter)


def _compute_index_and_weights(maparray, mapvalues):
    '''helper function  to get indices and weights for interpolating tecmaps


    Args:
        maparray (np.array) : array to get indices in
        mapvalues (Union[float,np.array]) :  values to get indices for
    Returns:
        Tuple[np.array, np.array, np.array]: idx1,idx2 and weights for idx2,
                                             idx2 is always >= idx1

    '''
    is_reverse = maparray[1] < maparray[0]
    idx1 = np.argmin(np.absolute(maparray[np.newaxis]
                                 - mapvalues[:, np.newaxis]), axis=1)
    idx2 = idx1.copy()
    if not is_reverse:
        idx1[maparray[idx1] > mapvalues] -= 1
        idx2[maparray[idx2] < mapvalues] += 1
    else:
        idx1[maparray[idx1] < mapvalues] -= 1
        idx2[maparray[idx2] > mapvalues] += 1
    idx1[idx1 < 0] = 0
    idx2[idx2 < 0] = 0
    idx1[idx1 >= maparray.shape[0]] = maparray.shape[0] - 1
    idx2[idx2 >= maparray.shape[0]] = maparray.shape[0] - 1
    _steps = np.absolute(maparray[idx2] - maparray[idx1])
    weights = np.absolute(mapvalues - maparray[idx1])
    weights[_steps == 0] = 1.
    weights[_steps!=0] = weights [_steps!=0]/ _steps[_steps!=0]
    return idx1, idx2, weights


def compute_tec_interpol(times, lats, lons, tecinfo, apply_earth_rotation=0):
    '''Get interpolated TEC for array of times/lats and lons

    Derive interpolated (4 point in lon,lat,2 point in time) vTEC values,
    optionally correcting for earth rotation.

    Args:
        lats (np.array) : angles in degrees between -90 and 90
        lons (np.array) : angles in degrees between -180,180
        times (np.array) : times in decimal hour of the day (eg. 23.5
        for half past 11PM)
        tecinfo (tuple) : tuple with the return values of read_tec function
        apply_earth_rotation(float) : specify (with a number between 0 and 1)
        how much of the earth rotaion is taken in to account in the
        interpolation step.
        This is assuming that the TEC maps move according to the rotation
        Earth (following method 3 of interpolation described in the IONEX
        document). Experiments with high time resolution ROB data show that
        this is not really the case, resulting in strange wavelike structures
        when applying this smart interpolation. Negative values of this
        parameter would result in an ionosphere that moves with the rotation
        of the earth
    Returns:
        np.array : interpolated tecvalues
    '''
    assert times.shape == lats.shape,\
        "times and lats should be array with same shape"
    assert times.shape == lons.shape, \
        "times and lons should be array with same shape"
    tecdata = tecinfo[0]  # TEC in TECU
    lonarray = tecinfo[2]  # longitude in degrees from West to East (- to +)
    latarray = tecinfo[3]  # lattitude in degrees
    # times in hour of the day (eg. 23.5 for half past 11PM)
    maptimes = tecinfo[4]

    # get indices of nearest 2 time frames + inverse distance weights
    # assume time is  sorted from early to late
    timeidx1, timeidx2, time_weights = _compute_index_and_weights(
        maptimes, times)

    # latarray is sorted small to large
    latidx1, latidx2, lat_weights = _compute_index_and_weights(
        latarray, lats)

    # for getting lon idx take into account earth_rotation
    # if longitudes cover all range between -180 and 180 you can modulate the
    # indices, otherwise we have to take the edge of the map.
    lonstep = lonarray[1] - lonarray[0]
    full_circle = np.remainder(lonarray[0] - lonarray[-1], 360.)\
        <= 1.1 * lonstep

    rot1 = ((times - maptimes[timeidx1]) * 360. / 24.) * apply_earth_rotation
    rot2 = ((times - maptimes[timeidx2]) * 360. / 24.) * apply_earth_rotation

    if not full_circle:
        lonidx11, lonidx12, lon_weights1 = _compute_index_and_weights(
            lonarray, lons + rot1)
        lonidx21, lonidx22, lon_weights2 = _compute_index_and_weights(
            lonarray, lons + rot2)
    else:
        lonidx11 = np.argmin(np.absolute(np.remainder(lonarray[np.newaxis]
                                                      - rot1[:, np.newaxis]
                                                      - lons[:, np.newaxis]
                                                      + 180., 360.)
                                         - 180.), axis=1)
        lonidx12 = lonidx11.copy()
        lonidx11[np.remainder(lonarray[lonidx11] - rot1 - lons + 180., 360.)
                 - 180. > 0] -= 1
        lonidx12[np.remainder(lonarray[lonidx12] - rot1 - lons + 180., 360.)
                 - 180. < 0] += 1
        lonidx11[lonidx11 < 0] += lonarray.shape[0]
        lonidx12[lonidx12 < 0] += lonarray.shape[0]
        lonidx11[lonidx11 >= lonarray.shape[0]] -= lonarray.shape[0]
        lonidx12[lonidx12 >= lonarray.shape[0]] -= lonarray.shape[0]
        lon_weights1 = np.absolute(np.remainder(lonarray[lonidx11]
                                                - rot1
                                                - lons + 180., 360.)
                                   - 180.) / lonstep

        lonidx21 = np.argmin(np.absolute(np.remainder(lonarray[np.newaxis]
                                                      - rot2[:, np.newaxis]
                                                      - lons[:, np.newaxis]
                                                      + 180., 360.)
                                         - 180.), axis=1)
        lonidx22 = lonidx21.copy()
        lonidx21[np.remainder(lonarray[lonidx21] - rot2 - lons + 180., 360.)
                 - 180. > 0] -= 1
        lonidx22[np.remainder(lonarray[lonidx22] - rot2 - lons + 180., 360.)
                 - 180. < 0] += 1
        lonidx21[lonidx21 < 0] += lonarray.shape[0]
        lonidx22[lonidx22 < 0] += lonarray.shape[0]
        lonidx21[lonidx21 >= lonarray.shape[0]] -= lonarray.shape[0]
        lonidx22[lonidx22 >= lonarray.shape[0]] -= lonarray.shape[0]
        lon_weights2 = np.absolute(np.remainder(lonarray[lonidx21]
                                                - rot2
                                                - lons + 180., 360.)
                                   - 180.) / lonstep
    logging.debug("inidces time %d %d indices lat %d %d indices \
                  lon %d %d %d %d", timeidx1[0], timeidx2[0],
                  latidx1[0], latidx2[0],
                  lonidx11[0], lonidx12[0],
                  lonidx21[0], lonidx22[0])
    logging.debug("weights time %f lat %f lon %f %f",
                  time_weights[0],
                  lat_weights[0],
                  lon_weights1[0],
                  lon_weights2[0])
    tecs = (tecdata[timeidx1, latidx1, lonidx11] * (1. - lon_weights1)
            + tecdata[timeidx1, latidx1, lonidx12] * lon_weights1) \
        * (1. - time_weights)
    tecs += (tecdata[timeidx2, latidx1, lonidx21] * (1. - lon_weights2)
             + tecdata[timeidx2, latidx1, lonidx22] * lon_weights2) \
        * (time_weights)
    tecs *= 1. - lat_weights
    tecs += lat_weights \
        * (tecdata[timeidx1, latidx2, lonidx11] * (1. - lon_weights1)
           + tecdata[timeidx1, latidx2, lonidx12] * lon_weights1) \
        * (1. - time_weights)
    tecs += lat_weights \
        * (tecdata[timeidx2, latidx2, lonidx21] * (1. - lon_weights2)
           + tecdata[timeidx2, latidx2, lonidx22] * lon_weights2) \
        * (time_weights)
    return tecs


def getTECinterpol(time, lat, lon, tecinfo, apply_earth_rotation=0):
    """old function name for compatibility. Use compute_tec_interpol
    instead"""

    #logging.warning("obsolete, use compute_tec_interpol instead")
    if np.isscalar(time):
        time = [time]
        lat = [lat]
        lon = [lon]
    return compute_tec_interpol(np.array(time), np.array(lat), np.array(lon),
                                tecinfo,
                                apply_earth_rotation)


def _combine_ionex(outpath, filenames, newfilename):
    """Helper function to combine separate IONEXfiles into 1 single file
    (needed for 15min ROBR data)"""

    if os.path.isfile(outpath + newfilename):
        logging.info("FILE exists: " + outpath + newfilename)
        return outpath + newfilename
    newf = open(outpath + newfilename, 'w')
    filenames = sorted(filenames)
    firstfile = open(filenames[0])
    lastfile = open(filenames[-1])
    for line in lastfile:
        if "EPOCH OF LAST MAP" in line:
            lastepoch = line
            lastfile.close()
            break
    # write header + tec map
    for line in firstfile:
        if "END OF TEC MAP" in line:
            newf.write(line)
            break
        if "EPOCH OF LAST MAP" not in line:
            if "OF MAPS IN FILE" in line:
                newf.write(line.replace('1', str(len(filenames))))
            else:
                newf.write(line)
        else:
            newf.write(lastepoch)
    tecmapnr = 2
    for myfname in filenames[1:]:
        myf = open(myfname)
        end_of_header = False
        for line in myf:
            if not end_of_header and "END OF HEADER" in line:
                end_of_header = True
            else:
                if end_of_header:
                    if "END OF TEC MAP" in line:
                        newf.write(line.replace('1', str(tecmapnr)))
                        break
                    if "START OF TEC MAP" in line:
                        newf.write(line.replace('1', str(tecmapnr)))
                    else:
                        newf.write(line)
        tecmapnr += 1
    newf.write("END OF FILE\n")
    return os.path.join(outpath, newfilename)
    # ignore RMS map for now, since it is filled with zeros anyway


def _gunzip_some_file(compressed_file,
                      uncompressed_file,
                      delete_file=1):
    command = "gunzip -dc %s > %s" % (compressed_file, uncompressed_file)
    retcode = os.system(command)
    if retcode:
        raise RuntimeError("Could not run '%s'" % command)
    if delete_file:
        os.remove(compressed_file)
    return uncompressed_file


def _store_files(ftp, filenames, outpath, overwrite=False):
    """helper function to store files from ftp server to outpath"""

    nfilenames = []
    for myf in filenames:
        #make sure filename is always stored uppercase
        mypath = os.path.join(outpath, myf.upper())
        if not overwrite and os.path.isfile(mypath):
            nfilenames.append(mypath)
            logging.info("file %s exists", mypath)
        elif not overwrite and\
        os.path.isfile(mypath.replace(".Z","")):
            nfilenames.append(mypath.replace(".Z",""))
            logging.info("file %s exists",
                         mypath.replace(".Z",""))
        else:
            myp = open(mypath, "wb")
            ftp.retrbinary("RETR " + myf, myp.write)
            myp.close()
            nfilenames.append(mypath)
    nfilenames_copy = nfilenames[:]
    nfilenames = []
    for myf in nfilenames_copy:
        if myf.endswith(".Z"):
            nfilenames.append(_gunzip_some_file(
                myf,
                myf.replace(".Z","")))
        else:
            nfilenames.append(myf)
    return nfilenames


def _get_IONEX_file(time="2012/03/23/02:20:10.01",
                    server="ftp://ftp.aiub.unibe.ch/CODE/",
                    prefix="codg",
                    outpath='./',
                    overwrite=False,
                    backupserver="ftp://ftp.aiub.unibe.ch/CODE/"):
    """Get IONEX file with prefix from server for a given day

    Downloads files with given prefix from the ftp server, unzips and stores
    the data. For prefix ROBR the data is stored on the server in a separate
    IONEX file every 15 minutes, these are automatically combined for
    compatibility.

    Args:
        time (string or list) : date of the observation
        server (string) : ftp server + path to the ionex directories
        prefix (string) : prefix of the IONEX files (case insensitive)
        outpath (string) : path where the data is stored
        overwrite (bool) : Do (not) overwrite existing data
    """
    prefix=prefix.upper()
    if outpath[-1] != "/":
        outpath += "/"
    if not os.path.isdir(outpath):
        try:
            os.makedirs(outpath)
        except:
            logging.error("cannot create output dir for IONEXdata: %s",
                          outpath)

    try:
        yy = time[2:4]
        year = int(time[:4])
        month = int(time[5:7])
        day = int(time[8:10])
    except:
        year = time[0]
        yy = year - 2000
        month = time[1]
        day = time[2]
    mydate = datetime.date(year, month, day)
    dayofyear = mydate.timetuple().tm_yday
    #if file exists just return filename
    if not overwrite and os.path.isfile("%s%s%03d0.%02dI"%(outpath,prefix,dayofyear,yy)):
        logging.info("FILE exists: %s%s%03d0.%02dI",outpath,prefix,dayofyear,yy)
        return "%s%s%03d0.%02dI"%(outpath,prefix,dayofyear,yy)
    #check if IGRG (fast files) exist, use those instead (UGLY!!)
    if not overwrite and os.path.isfile("%sIGRG%03d0.%02dI"%(outpath,dayofyear,yy)):
        logging.info("fast FILE exists: %sIGRG%03d0.%02dI",outpath,dayofyear,yy)
        return "%sIGRG%03d0.%02dI"%(outpath,dayofyear,yy)
   
    tried_backup=False
    serverfound=False
    while not serverfound: 
        ftpserver = server.replace("ftp:","").strip("/").split("/")[0]
        ftppath = "/".join(server.replace("ftp:","").strip("/").split("/")[1:])
        nr_tries = 0
        try_again = True
        while try_again and nr_tries<10:
            try:
                ftp = ftplib.FTP(ftpserver)
                ftp.login()
                try_again=False
                serverfound=True
            except ftplib.error_perm:
                if "213.184.6.172" in server:
                    ftp.login("data-out", "Qz8803#mhR4z")
                    try_again=False
                    serverfound=True
                else:
                    try_again=True
                    nr_tries += 1
                    if nr_tries>=10:
                        if tried_backup or server==backupserver:
                            raise Exception("Could not connect to %s"%ftpserver)
                        else:
                            server=backupserver
                            tried_backup=True
            except socket.gaierror:
                try_again=True
                nr_tries += 1
                if nr_tries>=10:
                    if tried_backup or server==backupserver:
                        raise Exception("Could not connect to %s"%ftpserver)
                    else:
                        server=backupserver
                        tried_backup=True
    ftp.cwd(ftppath)
    totpath = ftppath
    myl = []
    ftp.retrlines("NLST", myl.append)
    logging.info("Retrieving data for %d or %02d%03d", year, yy, dayofyear)
    if str(year) in myl:
        ftp.cwd(str(year))
        totpath += "/%d" % (year)
    elif "%02d%03d" % (yy, dayofyear) in myl:
        ftp.cwd("%02d%03d" % (yy, dayofyear))
        totpath += "/%02d%03d" % (yy, dayofyear)
    myl = []
    ftp.retrlines("NLST", myl.append)
    if "%03d"%dayofyear in myl:
        ftp.cwd("%03d"%dayofyear)
        totpath += "/%03d" % (dayofyear)
    logging.info("Retrieving data from %s", totpath)
    myl = []
    ftp.retrlines("NLST", myl.append)
    filenames = [i for i in myl if (prefix.lower() in i.lower()) and 
                 ("%03d"%dayofyear in i.lower()) and
                 (i.lower().endswith("i.z") or i.lower().endswith("i"))]
    logging.info(" ".join(filenames))
    #assert len(filenames) > 0, "No files found on %s for %s" % (server,prefix)
    if len(filenames) <=0:
        logging.info("No files found on %s for %s",server,prefix)
        return -1
        
    if prefix.lower() == "robr" and len(filenames) > 1:
        filenames = sorted(filenames)
        filenames = _store_files(ftp, filenames, outpath, overwrite)
        # get data for next day
        nextday = mydate + datetime.timedelta(days=1)
        nyear = nextday.year
        ndayofyear = nextday.timetuple().tm_yday
        ftp.cwd("/" + ftppath)
        myl = []
        ftp.retrlines("NLST", myl.append)
        if str(nyear) in myl:
            ftp.cwd(str(nyear))
        myl = ftp.retrlines("NLST")
        if str(ndayofyear) in myl:
            ftp.cwd(str(ndayofyear))
        myl = ftp.retrlines("NLST")
        nfilenames = [i for i in myl if (prefix.lower() in i.lower()) and
                      (i.lower().endswith("i.z")) and "A00" in i.upper()]
        nfilenames = _store_files(ftp, nfilenames, outpath, overwrite)
        filenames += nfilenames
        _combine_ionex(outpath, filenames,
                       prefix + "%03d0.%sI" % (dayofyear, yy))
        ftp.quit()
        return os.path.join(outpath, prefix + "%03d0.%sI" % (dayofyear, yy))
    else:
        nfilenames = _store_files(ftp, filenames, outpath, overwrite)
        ftp.quit()
        return nfilenames[0]

def get_urllib_IONEXfile(time="2012/03/23/02:20:10.01",
                    server="ftp://ftp.aiub.unibe.ch/CODE/",
                    prefix="codg",
                    outpath='./',
                    overwrite=False,
                    backupserver="ftp://ftp.aiub.unibe.ch/CODE/",
                    proxy_server=None,
                    proxy_type=None,
                    proxy_port=None,
                    proxy_user=None,
                    proxy_pass=None):
    """Get IONEX file with prefix from server for a given day

    Downloads files with given prefix from the ftp server, unzips and stores
    the data. Uses urllib2 instead of ftplib to have the option to use a ftp proxy server.

    Proxy args are optional.

    Args:
        time (string or list) : date of the observation
        server (string) : ftp server + path to the ionex directories
        prefix (string) : prefix of the IONEX files (case insensitive)
        outpath (string) : path where the data is stored
        overwrite (bool) : Do (not) overwrite existing data
        proxy_server (string): address of proxyserver, either url or ip address
        proxy_type (string): socks4 or socks5
        proxy_port (int): port of proxy server
        proxy_user (string): username for proxyserver
        proxy_pass (string): password for proxyserver
    """
    prefix=prefix.upper()
    if outpath[-1] != "/":
        outpath += "/"
    if not os.path.isdir(outpath):
        try:
            os.makedirs(outpath)
        except:
            print("cannot create output dir for IONEXdata: %s",
                          outpath)

    try:
        yy = int(time[2:4])
        year = int(time[:4])
        month = int(time[5:7])
        day = int(time[8:10])
    except:
        year = time[0]
        yy = year - 2000
        month = time[1]
        day = time[2]
    mydate = datetime.date(year, month, day)
    dayofyear = mydate.timetuple().tm_yday
    if not overwrite and os.path.isfile("%s%s%03d0.%02dI"%(outpath,prefix,dayofyear,yy)):
        logging.info("FILE exists: %s%s%03d0.%02dI",outpath,prefix,dayofyear,yy)
        return "%s%s%03d0.%02dI"%(outpath,prefix,dayofyear,yy)
    #check if IGRG (fast files) exist, use those instead (UGLY!!)
    if not overwrite and os.path.isfile("%sIGRG%03d0.%02dI"%(outpath,dayofyear,yy)):
        logging.info("fast FILE exists: %sIGRG%03d0.%02dI",outpath,dayofyear,yy)
        return "%sIGRG%03d0.%02dI"%(outpath,dayofyear,yy)
   
    tried_backup=False
    serverfound=False
    backupfound=False    
    #If proxy url is given, enable proxy using pysocks
    import urllib2
    if proxy_server and ("None" not in proxy_server):
        import socket
        import socks
        s = socks.socksocket()
        if proxy_type=="socks4":
            ProxyType = socks.SOCKS4
        if proxy_type=="socks5":
            ProxyType = socks.SOCKS5
        s.set_proxy(ProxyType, proxy_server, proxy_port, rdns=True, username=proxy_user, password=proxy_pass)

    # Url of the primary server has the syntax "ftp://ftp.aiub.unibe.ch/CODE/YYYY/CODGDOY0.YYI.Z" where DOY is the day of the year, padded with leading zero if <100, and YY is the last two digits of year.
    # Url of the backup server has the syntax "ftp://cddis.gsfc.nasa.gov/gnss/products/ionex/YYYY/DOY/codgDOY.YYi.Z where DOY is the day of the year, padded with leading zero if <100, and YY is the last two digits of year.
    #try primary url

    try:
        primary = urllib2.urlopen(server,timeout=30)
        serverfound = True
    except:
            try:
                secondary = urllib2.urlopen(backupserver,timeout=30)
                backupfound = True
                server=backupserver
            except:
                logging.error('Primary and Backup Server not responding') #enable in lover environment
    if "ftp://ftp.aiub.unibe.ch" in server:
        url = "ftp://ftp.aiub.unibe.ch/CODE/%4d/%s%03d0.%02dI.Z"%(year,prefix.upper(),dayofyear,yy)
    elif "ftp://cddis.gsfc.nasa.gov" in server:
        url = "ftp://cddis.gsfc.nasa.gov/gnss/products/ionex/%4d/%03d/%s%03d0.%02di.Z"%(year,dayofyear,prefix,dayofyear,yy)
    elif "igsiono.uwm.edu.pl" in server:
        url = "https://igsiono.uwm.edu.pl/data/ilt/%4d/igrg%03d0.%02di"%(year,dayofyear,yy)

    # Download IONEX file, make sure it is always uppercase
    fname = outpath+'/'+(url.split('/')[-1]).upper()
    try:
        site = urllib2.urlopen(url,timeout=30)
    except:
        logging.info("No files found on %s for %s",server,fname)
        return -1

    output=open(fname,'wb')
    output.write(site.read())
    output.close()
    ###### gunzip files
    if fname[-2:].upper()==".Z":
        command = "gunzip -dc %s > %s" % (fname, fname[:-2])
        retcode = os.system(command)
        if retcode:
            raise RuntimeError("Could not run '%s'" % command)
        else:
            os.remove(fname)
        fname=fname[:-2]
    #returns filename of uncompressed file
    return fname

def getIONEXfile(time="2012/03/23/02:20:10.01",
                 server="ftp://ftp.aiub.unibe.ch/CODE/",
                 prefix="codg",
                 outpath='./',
                 overwrite=False):
    getIONEXfile.__doc__ = _get_IONEX_file.__doc__
    return _get_IONEX_file(time, server, prefix, outpath, overwrite)

def get_TEC_data(times, lonlatpp, server, prefix, outpath, use_filter=None,earth_rot=0.):
    '''Returns vtec for given times and lonlats.
    If times has the same length as the first axis  of lonlatpp, 
    it is assumed that there is a one to one correspondence.
    Else vTEC is calculated for every combination of lonlatpp and times.
    
    Args:
        times (np.array) : float of time in MJD seconds
        lonlatpp (np.array) : array of time X 2 ,longitude lattitude of 
                              piercepoints
        server (string) : ftp server to get IONEX data from
        prefix (string) : prefix of IONEX data 
        outpath (string) : local location of the IONEX files
    Returns:
        np.array : array with shape times.shape x lonlatpp.shape[0], unless both
                   are equal
    '''
    
    date_parms = PosTools.obtain_observation_year_month_day_fraction(times[0])
    ionexf=get_IONEX_file(time=date_parms,server=server,prefix=prefix,outpath=outpath)
    tecinfo=ionex.readTEC(ionexf,use_filter=use_filter)
    latpp = lonlatpp[:, 1]
    lonpp = lonlatpp[:, 0]
    if latpp.shape == times.shape:
        vtec = compute_tec_interpol(times,lat=latpp,lon=lonpp,tecinfo=tecinfo,apply_earth_rotation=earth_rot)
    else:
        vtec=[]
        for itime in range(times.shape[0]):
            vtec.append(compute_tec_interpol(times[itime]*np.ones_like(latpp),
                                             lat=latpp,
                                             lon=lonpp,
                                             tecinfo=tecinfo,
                                             apply_earth_rotation=earth_rot))
    return np.array(vtec)
            



import sys
import RMextract.getIONEX as ionex
from RMextract import PosTools
import numpy as np
import logging
import argparse


def main():
    parser = argparse.ArgumentParser(description='Downloads relevant IONEX files for a given MS')
    parser.add_argument('MSfiles', type=str,
                            help='MS for which the parmdb should be created.')
    parser.add_argument('--server', type=str, default='ftp://ftp.aiub.unibe.ch/CODE/',
                            help='URL of the server to use. (default: ftp://ftp.aiub.unibe.ch/CODE/)')
    parser.add_argument('--prefix', type=str, default='CODG',
                            help='Prefix of the IONEX files. (default: \"CODG\")')
    parser.add_argument('--ionexpath', '--path', type=str, default='IONEXdata/',
                            help='Directory where to store the IONEX files. (default: \"IONEXdata/\")')
    args = parser.parse_args()

    MS=args.MSfiles
    server=args.server
    prefix=args.prefix
    (timerange,timestep,pointing,stat_names,stat_pos)=PosTools.getMSinfo(MS)
    
    start_time = timerange[0]
    end_time = timerange[1]
    timerange[0] = start_time - timestep
    timerange[1] = end_time + timestep
    times,timerange=PosTools.getIONEXtimerange(timerange,timestep)
    if len(times[-1])==0 or times[-1][-1]<timerange[1]:
        timestmp=list(times[-1])
        timestmp.append(timerange[1]) #add one extra step to make sure you have a value for all times in the MS in case timestep hase been changed
        times[-1]=np.array(timestmp)

    for time_array in times[::-1]: #check in reverse order, since datamight not be available for latest days
        starttime=time_array[0]
        date_parms =  PosTools.obtain_observation_year_month_day_fraction(starttime)
        if not "http" in server: #ftp server use ftplib
            ionexf=ionex.getIONEXfile(time=date_parms,server=server,prefix=prefix,outpath=args.ionexpath)
        else:
            ionexf=ionex.get_urllib_IONEXfile(time=date_parms,server=server,prefix=prefix,outpath=args.ionexpath)
        if ionexf==-1:
            if not "igsiono.uwm.edu.pl" in server:
                logging.info("cannot get IONEX data, try fast product server instead")
                server="https://igsiono.uwm.edu.pl"
                prefix="igrg"
                ionexf=ionex.get_urllib_IONEXfile(time=date_parms,server=server,prefix=prefix,outpath=args.ionexpath)
            if ionexf==-1:
                logging.error("IONEX data not available, even not from fast product server")
                return False
        
    return True




#!/usr/bin/env python
from RMextract import PosTools
from RMextract import getIONEX as ionex;
import os
import numpy as np
from datetime import date

from RMextract.EMM import EMM as EMM

ION_HEIGHT=PosTools.ION_HEIGHT
#####################  main processing function #####################
def getRM(MS=None,
           server="ftp://ftp.aiub.unibe.ch/CODE/",
           prefix='codg',
           ionexPath="IONEXdata/",
           earth_rot=0,
           timerange=0,
           use_azel = False,ha_limit=-1000,use_filter=None,
           **kwargs):
    '''optional arguments are:
    radec or pointing : [ra,dec] in radians, or if use_azel =True, az + el in radians,
    timestep in s, timerange = [start, end]in MJD seconds, 
    otherwise use start_time/end_time (casa timestring,eg. 2012/11/21/12:00:00  or 56252.5d  NEEDS PYRAP) ,
    stat_names = list of strings per station, 
    stat_positions = list of length 3 numpy arrays, containing station_position in ITRF meters.
    useEMM = boolean, use EMM for Earth magnetic field, otherwise WMM cooefficients will be used.
    out_file = string, if given the data points will be written to a text file.
    use_mean = True if you only want report for mean of station positions
    use_filter =standard deviation,or list of standard deviations (time,long, lat) to gaussian filter TEC data 
    TIME_OFFSET = float, offset time at start and end to ensure all needed values are calculated,
    Returns the (timegrid,timestep,TEC) where TEC is a dictionary containing 1 enumpyarray per station in stat_names. 
    If stat_names is not given, the station names will either be extracted from the MS or st1...stN '''

    #print ('earth_rot', earth_rot)
    #print ('timerange ', timerange)
    #print ('use_azel ', use_azel)
    stat_names=[]
    useEMM=False
    use_mean = False
    myobject = ''
    timestep=60
    pointing=[0.,0.5*np.pi]
    out_file=''
    stat_pos=[PosTools.posCS002]
    use_proxy = False
    #stat_names=['LOFAR_CS002']
    if not (MS is None):

      (timerange,timestep,pointing,stat_names,stat_pos)=PosTools.getMSinfo(MS);
    
    for key in kwargs.keys():
        if key=='use_mean':
            use_mean = kwargs[key]
            stat_pos_mean = False
        if key=='radec' or key=='pointing':
            pointing = kwargs[key]
        if key=='object':
            myobject = kwargs[key]
            #out_file = 'RMextract_report_' + myobject
        if key=='start_time':
            start_time=kwargs[key]
            time_in_sec = False
        if key=='end_time':
            end_time=kwargs[key]
            time_in_sec = False
        if key=='timestep':
            timestep=kwargs[key]
        if key=='stat_names':
            stat_names=kwargs[key]
        if key=='out_file':
            out_file=kwargs[key]
        if key=='TIME_OFFSET':
           TIME_OFFSET=kwargs[key]
        if key=='stat_positions':
            stat_pos=kwargs[key]
            if not stat_names:
                stat_names =['st%d'%(i+1) for i in range(len(stat_pos))]
        if key=='useEMM':
            useEMM=kwargs[key]
        if key=="proxy_server":		#Check to see if user wants to use a proxy for downloading IONEX files.
            use_proxy = True
            proxy_server=kwargs["proxy_server"]
            proxy_type=kwargs["proxy_type"]
            proxy_port=kwargs["proxy_port"]
            proxy_user=kwargs["proxy_user"]
            proxy_pass=kwargs["proxy_pass"]
      
    if timerange != 0:
      start_time = timerange[0]
      end_time = timerange[1]
      time_in_sec = True
      reference_time = start_time
      timerange[0] = start_time - timestep
      timerange[1] = end_time + timestep
      str_start_time=PosTools.obtain_observation_year_month_day_hms(reference_time)
    else:
        timerange,str_start_time,reference_time=PosTools.get_time_range(start_time,end_time,timestep,time_in_sec,0)
    if str_start_time==-1:
       return
        
    if useEMM:
        #print ("USING EMM for EarthMagnetic Field")
        emm= EMM.EMM()
    else:
        emm= EMM.WMM()
    
    times,timerange=PosTools.getIONEXtimerange(timerange,timestep)
    if len(times[-1])==0 or times[-1][-1]<timerange[1]:
        timestmp=list(times[-1])
        timestmp.append(timerange[1]) #add one extra step to make sure you have a value for all times in the MS in case timestep hase been changed
        times[-1]=np.array(timestmp)
    timegrid=np.array([])
    TECs={};
    Bs={}
    Bpars={}
    air_mass={};
    RMs={};
    azimuth={};
    elevation={};
    flags={}
    if len(out_file) and len(myobject):
        out_file = out_file + '_' + myobject
    else:
      if len(myobject):
        out_file = 'RMextract_report_' + myobject
# print header section of report
    if len(out_file):
       if os.path.exists(out_file):
         os.remove(out_file)
       log = open(out_file, 'a')
       log.write ('Observing %s\n' % myobject)
       if not (MS is None):
           log.write ('Using measurement set %s\n' % MS)
       if use_azel:
         log.write ('observing at a fixed azimuth and elevation \n')
       log.write ('station_positions %s \n' % stat_pos)
       log.write ('\n')

    for st in stat_names:
        azimuth[st] = []
        elevation[st] = []
        air_mass[st] = []
        Bs[st]=[]
        Bpars[st]=[]
        RMs[st]=[]
        TECs[st]=[]
        air_mass[st]=[]
        flags[st]=[]
    for time_array in times:
        #get RM per timeslot
        starttime=time_array[0]
        #print ("getting ionexfile for",starttime)
        date_parms =  PosTools.obtain_observation_year_month_day_fraction(starttime)
        dayofyear = date(date_parms[0],date_parms[1],date_parms[2]).timetuple().tm_yday  
        emm.date=date_parms[0]+float(dayofyear)/365.
        #get relevant ionex file
        if not use_proxy:
            if not "http" in server: #ftp server use ftplib
                ionexf=ionex.getIONEXfile(time=date_parms,server=server,prefix=prefix,outpath=ionexPath)
            else:
                ionexf=ionex.get_urllib_IONEXfile(time=date_parms,server=server,prefix=prefix,outpath=ionexPath)
        else:
                ionexf=ionex.get_urllib_IONEXfile(time=date_parms,server=server,prefix=prefix,outpath=ionexPath,proxy_server=proxy_server,proxy_type=proxy_type,proxy_port=proxy_port,proxy_user=proxy_user,proxy_pass=proxy_pass)

        assert (ionexf!=-1),"error getting ionex data"
           
        tecinfo=ionex.readTEC(ionexf,use_filter=use_filter)
        if use_mean:
          if not stat_pos_mean:
            stn_mean = stat_pos.mean(0)
            stat_pos = []
            stat_pos.append(stn_mean)
            stat_pos_mean = True
            #print ('stat_pos.mean', stn_mean)
        for station,position in  zip(stat_names,stat_pos):
          #print ('generating data for station ', station)
 
          
          for time in time_array:
            result =  PosTools.obtain_observation_year_month_day_fraction(time)
            part_of_day= result[3] * 24
            if use_azel:
              az = pointing[0]
              el = pointing[1]
              flags[station].append(1)   
            else:
               az,el = PosTools.getAzEl(pointing,time,position,ha_limit)
               if az==-1 and el==-1:
                  return
               if az==999 and el==999:
                    #outsite hadec range
                  air_mass[station].append(-999)
                  azimuth[station].append(-999)
                  elevation[station].append(-999)
                  Bs[station].append([-999,-999,-999])
                  Bpars[station].append(-999)
                  RMs[station].append(-999)
                  TECs[station].append(-999)  #sTEC value
                  flags[station].append(0)
                  continue
            flags[station].append(1)   
            latpp,lonpp,height,lon,lat,am1=PosTools.getlonlatheight(az,el,position)

            if latpp==-1 and lonpp==-1 and height==-1:
               return
            #get VTEC from IONEX interpolation
            vTEC=ionex.getTECinterpol(time=part_of_day,lat=latpp,lon=lonpp,tecinfo=tecinfo,apply_earth_rotation=earth_rot)

            TECs[station].append(vTEC*am1)  #STEC value
            emm.lon = lonpp
            emm.lat = latpp
            emm.h= ION_HEIGHT / 1.0e3
            Bpar=-1*emm.getProjectedField(lon,lat)# minus sign since the radiation is towards the Earth
            BField=emm.getXYZ()
            Bs[station].append(BField)

            air_mass[station].append(am1)
            azimuth[station].append(az)
            elevation[station].append(el)
            Bpars[station].append(Bpar)
            #calculate RM constant comes from VtEC in TECU,B in nT, RM = 2.62
            RMs[station].append((Bpar*vTEC*am1)*2.62e-6)
          if use_mean:
            break

            
        timegrid=np.concatenate((timegrid,time_array));
    
    for st in stat_names:
      air_mass[station] = np.array(air_mass[st])
      azimuth[station] = np.array(azimuth[st])
      elevation[station] = np.array(elevation[st])
      TECs[st]=np.array(TECs[st])
      Bs[st]=np.array(Bs[st])
      Bpars[st]=np.array(Bpars[st])
      RMs[st]=np.array(RMs[st]) 
      flags[st]=np.array(flags[st])
      if use_mean:
        break
       
    big_dict={}
    big_dict['STEC']=TECs
    big_dict['Bpar']=Bpars
    big_dict['BField']=Bs
    big_dict['AirMass']=air_mass
    big_dict['elev']=elevation
    big_dict['azimuth']=azimuth
    big_dict['RM']=RMs 
    big_dict['times']=timegrid
    big_dict['timestep']=timestep
    big_dict['station_names'] = stat_names
    big_dict['stat_pos'] = stat_pos
    big_dict['flags'] = flags
    big_dict['reference_time'] = reference_time 
    # finish writing computed data to report
    if len(out_file):
       time_range=[big_dict['times'][0],big_dict['times'][-1]]
       log.write ('start and end times %s %s \n' % (time_range[0], time_range[1]))
       log.write ('reference time for rel_time=0: year month day hr min sec %s %s %s %s %s %s \n' % str_start_time)
       if use_azel:
         log.write ('observing at azimuth and elevation %s %s \n' % (pointing[0], pointing[1]))
       else:
         log.write ('observation direction %s %s \n' % (pointing[0], pointing[1]))
       log.write ('\n')
       k = 0
       for key in big_dict['station_names']:
         seq_no = 0
         if use_mean:
           log.write ('data for station mean position at %s\n' % (stat_pos[k]))
         else:
           log.write ('data for station %s  at position %s\n' % (key, stat_pos[k]))
         log.write ('seq  rel_time time_width El         Az         STEC           RM (rad/m2)   VTEC factor  \n')
         for i in range (timegrid.shape[0]):
           el = big_dict['elev'][key][i]
           if el < 0 :
             ok = 1
             stec = 0.0
             rm = 0.0
             vtec_factor = 1.0
           else:
             ok = 0
             stec =big_dict['STEC'][key][i]
             rm = big_dict['RM'][key][i]
             vtec_factor = 1.0 / big_dict['AirMass'][key][i]
           az = big_dict['azimuth'][key][i]
           rel_time = timegrid[i] - reference_time
           if i  == 0:
             time_width = reference_time - timegrid[i] 
           else:
             time_width = timegrid[i] - timegrid[i-1]
           log.write("%s : %s %s %s %s %s %s %s %s\n" % (seq_no, ok, rel_time, time_width, el, az, stec, rm, vtec_factor))
           seq_no = seq_no + 1
         k = k + 1
         if use_mean:
           break
         log.write (' \n')
       log.close()
       print ('****** finished ionosphere predictions report: ', out_file)
    else:
      print ('*********** finished ionosphere predictions ***************')


    return big_dict





HAS_PYRAP = True
try:
  from pyrap import tables as tab
  import pyrap.quanta as qu
  from pyrap.measures import measures
  print ('pyrap will be used to compute positions')
except:
  print ('We will need PyEphem to perform calculations!')
  print ('the accuracy of results might decease a bit')
  HAS_PYRAP = False

HAS_EPHEM = True
try:
  import ephem
  if not HAS_PYRAP:
    print ('PyEphem will be used to perform calculations!')
except:
  HAS_EPHEM = False

from math import *
import math
import os
import numpy as np
import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord, EarthLocation, AltAz,Angle,ITRS,FK5
from scipy import interpolate

# height of thin layer ionosphere
ION_HEIGHT=450.e3;


R_earth=6364.62e3;
R_earthkm=6364.62;
earth_ellipsoid_a = 6378137.0;
earth_ellipsoid_a2 = earth_ellipsoid_a*earth_ellipsoid_a;
earth_ellipsoid_b = 6356752.3142;
earth_ellipsoid_b2 = earth_ellipsoid_b*earth_ellipsoid_b;
earth_ellipsoid_e2 = (earth_ellipsoid_a2 - earth_ellipsoid_b2) / earth_ellipsoid_a2;
posCS002=[3826577.1095  ,461022.900196, 5064892.758]
########################################
sm_a = 6378137.0
invf = 298.257223563
f = 1.0 / invf

# The following two functions were provided by Emil Lenc
# Convert latitude (radians), longitude (radians) and elevation (metres) 
# to ITRF XYZ and vice versa
def WGS84ToITRF(lat, lon, h): # WGS-84 to ITRF (input in radians)
        SINK = math.sin(lat)
        COSK = math.cos(lat)
        e2 = 2.0 * f - f * f
        v = sm_a / math.sqrt(1.0 - e2 * SINK * SINK)
        x = (v + h) * COSK * math.cos(lon)
        y = (v + h) * COSK * math.sin(lon)
        z = ((1 - e2) * v + h) * SINK
        return x, y, z

def ITRFToWGS84(x, y, z):
        e2 = 2.0 * f - f * f
        E = e2 / (1.0 - e2)
        b = sm_a * (1.0 - f)
        p = math.sqrt(x * x + y * y)
        q = math.atan2(z * sm_a, (p * b))
        lat = math.atan2((z + E * b * math.sin(q) * math.sin(q) * math.sin(q)), (p - e2 * sm_a * math.cos(q) * math.cos(q) * math.cos(q)))
        v = sm_a / math.sqrt(1.0 - e2 * math.sin(lat) * math.sin(lat))
        lon = math.atan2(y, x)
        h = (p / math.cos(lat)) - v
        lat = math.degrees(lat)
        lon = math.degrees(lon)
        return lat, lon, h       # output in degrees here

########################################
# convert geodetic latitude to geocentric latitude
# input and output in radians
def GeodeticToGeocentricLat(geodetic_lat, height):
  l_sin = math.sin(geodetic_lat)
  e2 = 2.0 * f - f * f
  div = math.sqrt(1 - e2 * l_sin**2)
  rn =  sm_a / div 
  rn_div = rn + height
  ratio = 1 - e2 * rn / rn_div
  tan_geocentric_lat = ratio * math.tan(geodetic_lat) 
  geocentric_lat = math.atan(tan_geocentric_lat)
  return geocentric_lat

########################################
# see http://stackoverflow.com/questions/15954978/ecef-from-azimuth-elevation-range-and-observer-lat-lon-alt

# input expected in degrees here
def aer2ecef(azimuthDeg, elevationDeg, slantRange, obs_lat, obs_lon, obs_alt):
    obs_lat_r = math.radians(obs_lat)
    obs_lon_r = math.radians(obs_lon)
    sitex, sitey, sitez = WGS84ToITRF(obs_lat_r,obs_lon_r,obs_alt)

    #some needed calculations
    slat = math.sin(math.radians(obs_lat))
    slon = math.sin(math.radians(obs_lon))
    clat = math.cos(math.radians(obs_lat))
    clon = math.cos(math.radians(obs_lon))

    azRad = math.radians(azimuthDeg)
    elRad = math.radians(elevationDeg)

    # az,el,range to sez convertion
    south  = -slantRange * math.cos(elRad) * math.cos(azRad)
    east   =  slantRange * math.cos(elRad) * math.sin(azRad)
    zenith =  slantRange * math.sin(elRad)


    x = ( slat * clon * south) + (-slon * east) + (clat * clon * zenith) + sitex
    y = ( slat * slon * south) + ( clon * east) + (clat * slon * zenith) + sitey
    z = (-clat *        south) + ( slat * zenith) + sitez
    lat,lon,h = ITRFToWGS84(x, y, z)
    return lat, lon, h  # data are in units of degrees

################ from JMA_tools in ALBUS ########################
def get_day_of_year(year, month, day):
    """Get the day-of-year integer from the year/month/day

    year   I  YYYY
    month  I  MM starting from January == 1
    day    I  DD starting from 1 == 1
    """
    day_of_year_list = [0,31,59,90,120,151,181,212,243,273,304,334]
    doy = day_of_year_list[month-1] + day
    if(month>2):
        if((year&0x3)==0):
            if((year % 100 != 0) or (year % 400 == 0)):
                doy = doy+1
    return doy
################ from JMA_tools in ALBUS ########################
def get_ymdf_from_JD(JD):
    """get the year, month, day, day_fraction from an MJD

Taken from _Astronomical Algorithms_, Meeus, 1991
    """
    JD2 = JD + 0.5
    Z = int(JD2)
    F = JD2 - Z
    A = Z
    if(Z>= 2299161):
        alpha = int((Z-1867216.25)/36524.25)
        A = Z + 1 + alpha - int(alpha/4)
    B = A + 1524
    C = int((B-122.1)/365.25)
    D = int(365.25*C)
    E = int((B-D)/30.6001)
    day_total = B - D - int(30.6001*E) + F
    month = E - 1
    if(E >= 14): month = E - 13
    year = C - 4716
    if(month <= 2): year += 1
    day = int(day_total)
    day_fraction = day_total - day
    return year, month, day, day_fraction

################################################################################
def get_hms_from_frac(day_fraction):
    """get hours, minues, seconds from a fractional day.

Does not worry about leap seconds.
"""
    h = day_fraction * 24.0
    hour = int(h+2E-13)
    m = (h - hour) * 60.0
    minute = int(m+1E-11)
    second = (m - minute) * 60.0
    return hour, minute, second


################################################################################
def get_ymdh_from_JD(JD):
    """get hours, minues, seconds from a fractional day.
"""
    year, month, day, day_fraction = get_ymdf_from_JD(JD)
    hour, minute, second = get_hms_from_frac(day_fraction)
    return year, month, day, hour, minute, second

################################################################################
def obtain_observation_year_month_day_fraction(start_time):
    julian_day = (start_time / 86400.0) + 2400000.5
    result = get_ymdf_from_JD(julian_day)
    return result

################################################################################
# Get the day of year from the Year, month, day for the start of observations
def obtain_observation_year_month_day_hms(start_time):
    if HAS_PYRAP:
      #print ("getting time", str(start_time)+'s')
      date_list = qu.quantity(str(start_time)+'s').formatted("YMD").split("/")
      year = int(date_list[0])
      month = int(date_list[1])
      day = int(date_list[2])
      time_list=date_list[3].split(":")
      return (year, month, day,int(time_list[0]),int(time_list[1]),float(time_list[2]))
    else:
      julian_day = (start_time / 86400.0) + 2400000.5
      year, month, day, hour, minute, second  = get_ymdh_from_JD(julian_day)
      return (year, month, day,hour, minute, second)





def getMSinfo(MS=None):
    if MS is None:
        print ("No measurement set given")
        return
    if not HAS_PYRAP:
        print ("Install pyrap to be able to extract info from MS")
        return

        
    if os.path.isdir(MS):
        myMS=tab.table(MS)
    else:
        print ("Do not understand the format of MS",MS,"bailing out")
        return;
    timerange=[np.amin(myMS.getcol('TIME_CENTROID')),np.amax(myMS.getcol('TIME_CENTROID'))]
    timestep=myMS.getcell('INTERVAL',0)
    
    pointing= tab.table(myMS.getkeyword('FIELD')).getcell('PHASE_DIR',0);    
    stations = tab.table(myMS.getkeyword('ANTENNA')).getcol('NAME')
    station_pos = tab.table(myMS.getkeyword('ANTENNA')).getcol('POSITION')

    return (timerange,timestep,pointing.flatten(),stations,station_pos)


def getPPsimple(height=[ION_HEIGHT,],mPosition=[0.,0.,0.],direction=[0.,0.,0.]):
    '''get piercepoints for antenna position mPosition in m, direction ITRF in m on unit sphere and for array of heights, assuming a spherical Earth'''
    height=np.array(height)
    stX=mPosition[0]
    stY=mPosition[1]
    stZ=mPosition[2]
    x=np.divide(stX,(R_earth+height))
    y=np.divide(stY,(R_earth+height))
    z=np.divide(stZ,(R_earth+height))
    
    c = x*x + y*y + z*z - 1.0;
   
    dx=np.divide(direction[0],(R_earth+height))
    dy=np.divide(direction[1],(R_earth+height))
    dz=np.divide(direction[2],(R_earth+height))

    a = dx*dx + dy*dy + dz*dz;
    b = x*dx + y*dy  + z*dz;

    alpha = (-b + np.sqrt(b*b - a*c))/a;


    pp=np.zeros(height.shape+(3,))
    pp[:,0]=stX+alpha*direction[0]
    pp[:,1]=stY+alpha*direction[1]
    pp[:,2]=stZ+alpha*direction[2]

    am=np.divide(1.,pp[:,0]*dx+pp[:,1]*dy+pp[:,2]*dz)
    return pp,am

def getPPsimpleAngle(height=[ION_HEIGHT,],mPosition=[0.,0.,0.],direction=[0.,0.,0.]):
    '''get (lon,lat,h values) of piercepoints for antenna position mPosition in m, direction ITRF in m on unit sphere and for array of heights, assuming a spherical Earth'''
    height=np.array(height)
    stX=mPosition[0]
    stY=mPosition[1]
    stZ=mPosition[2]
    x=np.divide(stX,(R_earth+height))
    y=np.divide(stY,(R_earth+height))
    z=np.divide(stZ,(R_earth+height))
    
    c = x*x + y*y + z*z - 1.0;
   
    dx=np.divide(direction[0],(R_earth+height))
    dy=np.divide(direction[1],(R_earth+height))
    dz=np.divide(direction[2],(R_earth+height))

    a = dx*dx + dy*dy + dz*dz;
    b = x*dx + y*dy  + z*dz;

    alpha = (-b + np.sqrt(b*b - a*c))/a;


    pp=np.zeros(height.shape+(3,))
    pp[:,0]=stX+alpha*direction[0]
    pp[:,1]=stY+alpha*direction[1]
    pp[:,2]=stZ+alpha*direction[2]

    am=np.divide(1.,pp[:,0]*dx+pp[:,1]*dy+pp[:,2]*dz)

    ppl=np.zeros(height.shape+(3,))
    ppl[:,0]=np.arctan2(pp[:,1],pp[:,0])
    ppl[:,1]=np.arctan2(pp[:,2],np.sqrt(pp[:,0]*pp[:,0]+pp[:,1]*pp[:,1]))
    ppl[:,2]=height

    return ppl,am
    

def getPP(h=ION_HEIGHT,mPosition=[0.,0.,0.],direction=[0.,0.,0.]):
    stationX = mPosition[0];
    stationY = mPosition[1];
    stationZ = mPosition[2];

    ion_ellipsoid_a = earth_ellipsoid_a + h;
    ion_ellipsoid_a2_inv = 1.0 / (ion_ellipsoid_a * ion_ellipsoid_a);
    ion_ellipsoid_b = earth_ellipsoid_b + h;
    ion_ellipsoid_b2_inv = 1.0 / (ion_ellipsoid_b * ion_ellipsoid_b);
    
    x = stationX/ion_ellipsoid_a;
    y = stationY/ion_ellipsoid_a;
    z = stationZ/ion_ellipsoid_b;
    c = x*x + y*y + z*z - 1.0;

    dx = direction [0]/ ion_ellipsoid_a;
    dy = direction [1] / ion_ellipsoid_a;
    dz = direction [2] / ion_ellipsoid_b;

    a = dx*dx + dy*dy + dz*dz;
    b = x*dx + y*dy  + z*dz;
    alpha = (-b + sqrt(b*b - a*c))/a;
    pp_x = stationX + alpha*direction[0];
    pp_y = stationY + alpha*direction[1]
    pp_z = stationZ + alpha*direction[2];

    normal_x = pp_x * ion_ellipsoid_a2_inv;
    normal_y = pp_y * ion_ellipsoid_a2_inv;
    normal_z = pp_z * ion_ellipsoid_b2_inv;
    norm_normal2 = normal_x*normal_x + normal_y*normal_y + normal_z*normal_z;
    norm_normal = sqrt(norm_normal2);
    sin_lat2 = normal_z*normal_z / norm_normal2;

 
    g = 1.0 - earth_ellipsoid_e2*sin_lat2;
    sqrt_g = sqrt(g);

    M = earth_ellipsoid_b2 / ( earth_ellipsoid_a * g * sqrt_g );
    N = earth_ellipsoid_a / sqrt_g;

    local_ion_ellipsoid_e2 = (M-N) / ((M+h)*sin_lat2 - N - h);
    local_ion_ellipsoid_a = (N+h) * sqrt(1.0 - local_ion_ellipsoid_e2*sin_lat2);
    local_ion_ellipsoid_b = local_ion_ellipsoid_a*sqrt(1.0 - local_ion_ellipsoid_e2);

    z_offset = ((1.0-earth_ellipsoid_e2)*N + h - (1.0-local_ion_ellipsoid_e2)*(N+h)) * sqrt(sin_lat2);

    x1 = stationX/local_ion_ellipsoid_a;
    y1 = stationY/local_ion_ellipsoid_a;
    z1 = (stationZ-z_offset)/local_ion_ellipsoid_b;
    c1 = x1*x1 + y1*y1 + z1*z1 - 1.0;

    dx = direction[0] / local_ion_ellipsoid_a;
    dy = direction[1] / local_ion_ellipsoid_a;
    dz = direction[2] / local_ion_ellipsoid_b;
    a = dx*dx + dy*dy + dz*dz;
    b = x1*dx + y1*dy  + z1*dz;
    alpha = (-b + sqrt(b*b - a*c1))/a;

    pp_x = stationX + alpha*direction[0];
    pp_y = stationY + alpha*direction[1]
    pp_z = stationZ + alpha*direction[2];

    normal_x = pp_x / (local_ion_ellipsoid_a * local_ion_ellipsoid_a);
    normal_y = pp_y / (local_ion_ellipsoid_a * local_ion_ellipsoid_a);
    normal_z = (pp_z-z_offset) / (local_ion_ellipsoid_b * local_ion_ellipsoid_b);

    norm_normal2 = normal_x*normal_x + normal_y*normal_y + normal_z*normal_z;
    norm_normal = sqrt(norm_normal2);
    
    pp_airmass = norm_normal / (direction[0]*normal_x + direction[1]*normal_y + direction[2]*normal_z);

    return (pp_x,pp_y,pp_z,pp_airmass)


def getLonLat(pos):
    #converts ITRF pos in xyz to lon lat 
    me=measures()
    a=me.measure(me.position('ITRF',str(pos[0])+'m',str(pos[1])+'m',str(pos[2])+'m'),"ITRF");
    return (a['m0']['value'],a['m1']['value'])

def getLonLatStation(az=0,el=0,pos=posCS002):
    #gets converts local station direction to ITRF lon/lat
    if not isinstance(az,str):
        az=str(az)+'rad';
    if not isinstance(el,str):
        el=str(el)+'rad';
    me=measures()
    me.do_frame(me.position('ITRF',str(pos[0])+'m',str(pos[1])+'m',str(pos[2])+'m'))
    #me.do_frame(me.epoch('utc', 'today'))
    direction=me.direction("AZELGEO",az,el)
    return me.measure(direction,"ITRF");


def radec2azel(ra,dec,time, pos):
    me=measures();
    if type(ra)!=str:
        ra=str(ra)+'rad';
    if type(dec)!=str:
        dec=str(dec)+'rad';
    phasedir=me.direction('J2000',ra,dec)
    t=me.epoch("UTC",qu.quantity(time));
    me.do_frame(t);

    p = me.position('ITRF',str(pos[0])+'m',str(pos[1])+'m',str(pos[2])+'m')
    me.do_frame(p);
    #print ("input radec2azel",phasedir,ra,dec,p,t)
    azel = me.measure(phasedir,'AZELGEO');
    return azel;

def azel2radec(az,el,time, pos):
    me=measures();
    if type(az)!=str:
        az=str(az)+'rad';
    if type(el)!=str:
        el=str(el)+'rad';
    phasedir=me.direction('AZELGEO',az,el)
    t=me.epoch("UTC",qu.quantity(time));
    me.do_frame(t);

    p = me.position('ITRF',str(pos[0])+'m',str(pos[1])+'m',str(pos[2])+'m')
    me.do_frame(p);

    radec = me.measure(phasedir,'RADEC');
    return radec;


def getuvw(ra,dec,time, pos1,pos2):
    me=measures();
    if type(ra)!=str:
        ra=str(ra)+'rad';
    if type(dec)!=str:
        dec=str(dec)+'rad';
    phasedir=me.direction('J2000',ra,dec)
    me.do_frame(phasedir);
    t=me.epoch("UTC",qu.quantity(time));
    me.do_frame(t);
    
    p = me.position('ITRF',str(pos1[0])+'m',str(pos1[1])+'m',str(pos1[2])+'m')
    bl = me.baseline('ITRF',str(pos2[0]-pos1[0])+'m',str(pos2[1]-pos1[1])+'m',str(pos2[2]-pos1[2])+'m')
    #print (bl)
    me.do_frame(p);
    #print (me.to_uvw(bl)['xyz'])
    #return uvw;
     
def getIONEXtimerange(timerange,timestep):
    #IONEX files go per day, check if more than one file is  needed.
    times=[];
    oldtimerange=-100
    while timerange[0]< timerange[1] and timerange[0]>oldtimerange:
      oldtimerange=timerange[0]
      #print (timerange)
      result =  obtain_observation_year_month_day_fraction(timerange[0])
      part_of_day = result[3]
      result2 =  obtain_observation_year_month_day_fraction(timerange[1])
      if result2[2]==result[2]:  #sameday
        times.append(np.arange(timerange[0],timerange[1]+timestep,timestep)) #make sure to include the last timestep
      else:
        nr_remaining_seconds=(1.-part_of_day) * 24.0 * 60. * 60.;
        #print ("new day",nr_remaining_seconds)
        times.append(np.arange(timerange[0],timerange[0]+nr_remaining_seconds,timestep));
      #print ("in postools",len(times),times[-1],timerange,nr_remaining_seconds,result2,result,times[-1][-1],part_of_day)
      if len(times[-1]):
        timerange[0]=times[-1][-1]+timestep
    return times,timerange

def getlonlatheight(az,el,position,h=ION_HEIGHT):
    if HAS_PYRAP:
      lonlat=getLonLatStation(az,el,pos=position);

      lon=lonlat['m0']['value'];
      lat=lonlat['m1']['value'];
      # convert to itrf coordinates on sphere with radius 1
      diritrf=[cos(lat)*cos(lon),cos(lat)*sin(lon),sin(lat)]
      # calculate piercepoint in xyz(code from Bas vd Tol)

      pos_pp = (ppx1,ppy1,ppz1,am1)=getPP(h=h,mPosition=position,direction=diritrf)

      #get pp in lon,lat h
      me=measures();
      pp1position=me.position("ITRF",str(ppx1)+'m',str(ppy1)+'m',str(ppz1)+'m')

      lonpp = degrees(pp1position['m0']['value']);
      latpp = degrees(pp1position['m1']['value']);
      height = pp1position['m2']['value'] / 1.0e3
    elif HAS_EPHEM:
      slantRange = 3.0e20    # seem to need a large value to
                               # get near the WNB measures equivalent
                               #             lat,lon,ht = aer2ecef(math.degrees(body.az), math.degrees(body.alt), slantRange, math.degrees(geocentric_latitude), location_lon, ION_HEIGHT)
      location_lat, location_lon, location_height = ITRFToWGS84(position[0], position[1], position[2])
      lat,lon,ht = aer2ecef(math.degrees(az), math.degrees(el), slantRange, location_lat, location_lon, ION_HEIGHT)
      # convert to itrf coordinates on sphere with radius 1
      lat = math.radians(lat)
      lon = math.radians(lon)
      diritrf=[cos(lat)*cos(lon),cos(lat)*sin(lon),sin(lat)]
      # calculate piercepoint in xyz(code from Bas vd Tol)

      pos_pp = (ppx1,ppy1,ppz1,am1)=getPP(h=ION_HEIGHT,mPosition=position,direction=diritrf)
      #get pp in lon,lat h
      latpp, lonpp, height =  ITRFToWGS84(ppx1,ppy1,ppz1)
    else:
      print ('unable to compute position parameters - exiting!')
      return -1,-1,-1
    return latpp, lonpp, height,lon,lat,am1

def far_to_near(altaz_far, distance, obstime=None):
    """Add distance to AltAz instance"""
    if obstime is None:
        obstime = altaz_far.obstime
    return AltAz(
        location=altaz_far.location,
        obstime=obstime,
        az=altaz_far.az,
        alt=altaz_far.alt,
        distance=distance,
    )

def distance_from_height(obs_altaz, height_start=50., height_end = 20200., height_unit = u.km ):
    """From an AltAz pointing, find the distance to where it crosses a plane 100 km above Earth"""
    try_distances = np.linspace(height_start*0.7,2*obs_altaz.secz*height_end) * height_unit
    dummytime = Time.now()
    try_altaz = far_to_near(obs_altaz, try_distances, obstime=dummytime)
    try_heights = EarthLocation(
        *(try_altaz.transform_to(ITRS(obstime=dummytime)).cartesian.xyz)
    )
    f_height_to_distance = interpolate.interp1d(
        try_heights.height.to(u.km), try_distances.to(u.km)
    )
    f_dist_to_latpp = interpolate.interp1d(
        try_distances.to(u.km), try_heights.lat.to(u.deg)
    )
    f_dist_to_lonpp = interpolate.interp1d(
        try_distances.to(u.km), try_heights.lon.to(u.deg)
    )
    f_dist_to_Rlocal = interpolate.interp1d(
        try_distances.to(u.km), try_heights.to(u.km).itrs.cartesian.norm()
    )
    return f_height_to_distance,f_dist_to_latpp,f_dist_to_lonpp, f_dist_to_Rlocal

def getProfile(source_pos, stat_pos, time, h = np.arange(60,20200,10)*u.km):
  '''Get profile of piercepoints through the atmosphere, given a station and source direction
  
  Args: 
  source_pos : SkyCoord or list(2) Ra,Dec J200
  stat_pos : EarthLocation or array(3) in m
  time : Time or float (mjd) (can be an array)
  h : array of heights

  Returns:
  Tuple[ np.array, np.array, np.array, float, float, np.array ]:
  

  '''
  # first calculate Az,El
  if not type(stat_pos) is EarthLocation:
    stat_pos = EarthLocation.from_geocentric(*list(stat_pos),unit=u.m)
  if not type(source_pos) is SkyCoord:
    source_pos = SkyCoord(source_pos[0],source_pos[1],unit=(u.hourangle,u.deg),frame = FK5)
  if not type(time) is Time:
    time = Time(time,format = 'mjd')
  aa = AltAz(location = stat_pos, obstime = time)
  azel = source_pos.transform_to(aa)
  direction = azel.transform_to(ITRS)
  londir = np.arctan2(direction.y.value,direction.x.value)
  latdir = np.arcsin(direction.z.value)
  
  fh,flat,flon,fRloc = distance_from_height(azel,h[0].to(u.km).value,h[-1].to(u.km).value, u.km)
  hdist = fh(h.to(u.km))
  latpp = flat(hdist)
  lonpp = flon(hdist)
  R_local = stat_pos.to(u.km).itrs.cartesian.norm().value
  Rpp = fRloc(hdist)
  #am = 1./((R_earthkm**2-hdist**2-(R_earthkm+h.to(u.km).value)**2)/(-2.*hdist*(R_earthkm+h.to(u.km).value)))
  am = 1./((R_local**2-hdist**2-Rpp**2)/(-2.*hdist*Rpp))
  return latpp, lonpp, latdir, londir, am
  
  

def getAzEl(pointing,time,position,ha_limit=-1000):

    if HAS_PYRAP:
        if ha_limit==-1000:
            azel=radec2azel(pointing[0],pointing[1],time=str(time)+'s',pos=position);
            az=azel['m0']['value']
            el=azel['m1']['value']
        else:
            me=measures()
            p=me.position("ITRF",str(position[0])+'m',str(position[1])+'m',str(position[2])+'m')
            t=me.epoch("UTC",qu.quantity(str(time)+'s'))
            phasedir=me.direction('J2000',str(pointing[0])+'rad',str(pointing[1])+'rad')
            me.doframe(p)
            me.doframe(t)
            hadec=me.measure(phasedir,"HADEC")
            if abs(hadec['m0']['value'])>ha_limit:
                print ("below horizon",tab.taql('calc ctod($time s)')[0],degrees(hadec['m0']['value']),degrees(hadec['m1']['value']))
                return 999,999
            else:
                azel=me.measure(phasedir,"AZELGEO")
  
                az=azel['m0']['value'];
                el=azel['m1']['value'];
    elif HAS_EPHEM:
        if ha_limit!=-1000:
            print ("limiting on HA/DEC not implemented for PyEphem yet, ignoring")
        location_lat, location_lon, location_height = ITRFToWGS84(position[0], position[1], position[2])
        location = ephem.Observer()
        # convert geodetic latitude to geocentric
        # flattening, f, defined above for WGS84 stuff
        geodet_lat = math.radians(location_lat)
        tan_geocentric_latitude =  math.tan(geodet_lat) * (1 - f) **2
        geocentric_latitude = GeodeticToGeocentricLat(geodet_lat, location_height)
        location.lat = geocentric_latitude
        location.lon = math.radians(location_lon)
        location.elevation = location_height
        location.pressure = 0.0
        #  convert to Dublin Julian Date for PyEphem
        location.date =  time/86400.0 - 15019.5
        lst = location.sidereal_time()
        equatorial = ephem.Equatorial(str(12.0 * math.degrees(pointing[0])/180),str(math.degrees(pointing[1])))
        body = ephem.FixedBody()
        body._ra = equatorial.ra
        body._dec = equatorial.dec
        body._epoch = equatorial.epoch
        body.compute(location)
        az = math.degrees(body.az)   * math.pi / 180.0
        el = math.degrees(body.alt) * math.pi / 180.0
    else: 
      print ('failure to get azimuth and elevation! Exiting!')
      return -1,-1
    return az,el


def get_time_range(start_time,end_time,timestep,time_in_sec,TIME_OFFSET=0):
    if HAS_PYRAP:
      try:
        start_time = qu.quantity(start_time).get_value()
        end_time = qu.quantity(end_time).get_value()
        print ('**** specified start and end time ', start_time, end_time)
        reference_time = start_time * 86400.0 - TIME_OFFSET
        st = reference_time - timestep 
        et = end_time * 86400.0 +  timestep + TIME_OFFSET
        #print ("getting string",reference_time)
        str_start_time =  obtain_observation_year_month_day_hms(reference_time)
        timerange= [st, et]
      except:
        print ('no time range given')
        print ('exiting')
        return -1,-1,-1
    elif HAS_EPHEM:
      if time_in_sec:
        dublin_start = start_time / 86400.0 -15019.5
        dublin_end = end_time / 86400.0 -15019.5
        start_time = ephem.julian_date(ephem.Date(dublin_start)) - 2400000.5
        end_time = ephem.julian_date(ephem.Date(dublin_end)) - 2400000.5
      else:
        start_time = ephem.julian_date(ephem.Date(start_time)) - 2400000.5
        end_time = ephem.julian_date(ephem.Date(end_time)) - 2400000.5
      print ('ephem start and end time ', start_time, end_time)
      reference_time = start_time * 86400.0 - TIME_OFFSET
      st = reference_time - timestep 
      et = end_time * 86400.0 + timestep + TIME_OFFSET
      str_start_time =  obtain_observation_year_month_day_hms(reference_time)
      timerange= [st, et]
    else:
      print ('unable to get time range so exiting!')
      return -1,-1,-1
    return timerange,str_start_time,reference_time



import RMextract.PosTools
import RMextract.getIONEX as ionex
import os
import numpy as np
from math import *

ION_HEIGHT=450.e3

def getTEC(MS=None,
           server="ftp://cddis.gsfc.nasa.gov/gnss/products/ionex/",
           prefix='codg',
           ionexPath="IONEXdata/",
           earth_rot=0,ha_limit=-1000,
           **kwargs):
    '''optional arguments are  radec or pointing : [ra,dec] in radians,
    azel : [az,el] in radians, If azel is specified, radec will be ignored
    timestep in s, timerange = [start, end]in MJD seconds, 
    otherwise use start_time/end_time (either MJD or casa timestring (NEEDS PYRAP), eg. 2012/11/21/12:00:00  or 56252.5d  ),    
    stat_names = list of strings per station, 
    stat_positions = list of length 3 numpy arrays, containing station_position in ITRF meters.
    Returns the (timegrid,timestep,TEC) where TEC is a dictionary containing 1 enumpyarray per station in stat_names. 
    If stat_names is not given, the station names will either be extracted from the MS or st1...stN '''

    stat_names=[]
    stat_pos=[PosTools.posCS002]
    #stat_names=['LOFAR_CS002']
    # the following parameter is used to extend the observation range by 120 sec
    # before and after the actual specified time. If we want to correct an
    # actual data set, this is required for the scipy 1d interpolation routine to
    # ensure that the calculated range of data exceeds the time range actually
    # observed - I have used the same value in ALBUS - Tony
    TIME_OFFSET = 0
    if not (MS is None):

        (timerange,timestep,pointing,stat_names,stat_pos)=PosTools.getMSinfo(MS);
    useAzel=False
    for key in kwargs.keys():
        if not useAzel and (key=='radec' or key=='pointing'):
            pointing = kwargs[key]
        if key=='start_time':
            start_time=kwargs[key]
            time_in_sec = False
        if key=='end_time':
            end_time=kwargs[key]
            time_in_sec = False
        if key=='azel':
            pointing = kwargs[key]
            useAzel=True
        if key=='timestep':
            timestep=kwargs[key]
        if key=='timerange':
            timerange=kwargs[key]
        if key=='stat_names':
            stat_names=kwargs[key]
        if key=='TIME_OFFSET':
           TIME_OFFSET=kwargs[key]
        if key=='stat_positions':
            stat_pos=kwargs[key]
            if not stat_names:
                stat_names =['st%d'%(i+1) for i in range(len(stat_pos))]

    if timerange != 0:
      start_time = timerange[0]-TIME_OFFSET
      end_time = timerange[1]+TIME_OFFSET
      time_in_sec = True
      reference_time = start_time
      str_start_time=PosTools.obtain_observation_year_month_day_hms(reference_time)
    else:
        timerange,str_start_time,reference_time=PosTools.get_time_range(start_time,end_time,timestep,time_in_sec,TIME_OFFSET)
    if str_start_time==-1:
       return

    #IONEX files go per day, check if more than one file is  needed.
    times,timerange=PosTools.getIONEXtimerange(timerange,timestep)
    #times.append(np.arange(timerange[0],timerange[1]+timestep,timestep)) #add one extra step to make sure you have a value for all times in the MS in case timestep hase been changed
    timegrid=np.array([])
    TECs={};
    ams={};
    flags={}
    for st in stat_names:
        TECs[st]=[]
        ams[st]=[]
        flags[st]=[]
    for time_array in times:
        #get RM per timeslot
        starttime=time_array[0];
        date_parms =  PosTools.obtain_observation_year_month_day_fraction(starttime)
        #get relevant ionex file
        ionexf=ionex.getIONEXfile(time=date_parms,server=server,prefix=prefix,outpath=ionexPath)
        if ionexf==-1:
            print ("error opening ionex data")
            return
       
        tecinfo=ionex.readTEC(ionexf)
        for station,position in  zip(stat_names,stat_pos):
          #print ("getting TEC for",station,"at",position,"time",time_array )
          for time in time_array:
            result =  PosTools.obtain_observation_year_month_day_fraction(time)
            part_of_day= result[3] * 24
            if useAzel:
                  az=pointing[0]
                  #el=0.5*np.pi-pointing[1]
                  el=pointing[1]
                  flags[station].append(1)
            else:

                az,el=PosTools.getAzEl(pointing,time,position,ha_limit)
                if az==-1 and el==-1:
                    return
                if az==999 and el==999:
                    #outsite hadec range
                    TECs[station].append(-999)  #sTEC value
                    ams[station].append(-999)
                    flags[station].append(0)
                    continue
                flags[station].append(1) 
            latpp,lonpp,height,lon,lat,am1=PosTools.getlonlatheight(az,el,position)
            if latpp==-1 and lonpp==-1 and height==-1:
               return
            vTEC=ionex.getTECinterpol(time=part_of_day,lat=latpp,lon=lonpp,tecinfo=tecinfo,apply_earth_rotation=earth_rot)

            TECs[station].append(vTEC)  #sTEC value
            ams[station].append(am1)
        timegrid=np.concatenate((timegrid,time_array));
    
    for st in stat_names:
            TECs[st]=np.array(TECs[st])
            ams[st]=np.array(ams[st])
            flags[st]=np.array(flags[st])
    return (timegrid,timestep,TECs,ams,flags)
        





"""
Extract RM values for a LOFAR observation. Fill h5parm

Created on Tue Aug 7 11:46:57 2018

@author: mevius

Changes for prefactor included by Alexander Drabent (02 June 2021)
"""
from losoto.h5parm import h5parm
from RMextract import getRM
from RMextract import PosTools
import RMextract.getIONEX as ionex
import pyrap.tables as pt
import os
import numpy as np
import sys
import logging
import argparse


def makesolset(MS, data, solset_name):
    solset = data.makeSolset(solset_name)    

    logging.info('Collecting information from the ANTENNA table.')
    antennaTable = pt.table(MS + "::ANTENNA", ack=False)
    antennaNames = antennaTable.getcol('NAME')
    antennaPositions = antennaTable.getcol('POSITION')
    antennaTable.close()
    antennaTable = solset.obj._f_get_child('antenna')
    antennaTable.append(list(zip(*(antennaNames,antennaPositions))))
    
    logging.info('Collecting information from the FIELD table.')
    fieldTable = pt.table(MS + "::FIELD", ack=False)
    phaseDir = fieldTable.getcol('PHASE_DIR')
    pointing = phaseDir[0, 0, :]
    fieldTable.close()

    sourceTable = solset.obj._f_get_child('source')
    # add the field centre, that is also the direction for Gain and Common*
    sourceTable.append([('pointing',pointing)])


def createRMh5parm(MSfiles, h5parmdb, solset_name = "sol000",all_stations=True,timestepRM=300,
                ionex_server="ftp://ftp.aiub.unibe.ch/CODE/",
                ionex_prefix='CODG',ionexPath="./",earth_rot=0,proxyServer=None,proxyPort=None,proxyType=None,proxyUser=None,proxyPass=None):
    '''Add rotation measure to existing h5parmdb
    
    Args:
        MSfiles (string) :  path + filename of Measurement Set 
        h5parmdb (string) : name of existing h5parm
        solset_name (string) : optional name of solset in h5parmdb, 
                            if not set, first one will be chosen
        all_stations (bool) : optional calculate RM values for all stations in the MS,'
                            default only for position of CS002LBA 
        timestep (float) : timestep in seconds
        ionex_server (string) : ftp server for IONEX files
        ionex_prefix (string) : prefix of IONEX files
        ionexPath (string) : location where IONEX files are stored
        earth_rot (float) : parameter to determine how much earth rotation is taken \
        into account when interpolating IONEX data. (0 is none, 1 is full)
        proxyserver (str): Name of a proxy server to use
        proxyport (int): Port of the proxy server to use
        proxytype (str): Type of the proxy server to use
        proxyuser (str): Username of the proxy server to use
        proxypass (str): Password of the proxy server to use
    '''
    
    try:
        mslist = MSfiles.lstrip('[').rstrip(']').replace(' ','').replace("'","").split(',')
    except AttributeError:
        mslist = MSfiles
    
    if len(mslist) == 0:
        raise ValueError("Did not find any existing directory in input MS list!")
        pass
    else:
        MS = mslist[0]
        pass
    
    if not os.path.exists(h5parmdb):
        logging.error('Could not find h5parmdb: ' + h5parmdb)
        return(1)
    
    data          = h5parm(h5parmdb, readonly=False)
    if not solset_name in data.getSolsetNames():
        makesolset(MS,data,solset_name)
    solset        = data.getSolset(solset_name)
    soltabs       = solset.getSoltabs()
    station_names = sorted(solset.getAnt().keys())

    if 'RMextract' in [s.name for s in soltabs]:
        logging.warning('Soltab RMextract exists already. Skipping...')
        return(0)
    
    #extract the timerange information
    (timerange,timestep,pointing,stat_names,stat_pos) = PosTools.getMSinfo(MS)
    start_time = timerange[0]
    end_time = timerange[1]
    timerange[0] = start_time - timestep
    timerange[1] = end_time + timestep
    times,timerange = PosTools.getIONEXtimerange(timerange, timestep)
    if len(times[-1]) == 0 or times[-1][-1] < timerange[1]:
        timestmp = list(times[-1])
        timestmp.append(timerange[1]) #add one extra step to make sure you have a value for all times in the MS in case timestep hase been changed
        times[-1] = np.array(timestmp)

    for time_array in times[::-1]:    #check in reverse order, since datamight not be available for latest days
        starttime = time_array[0]
        date_parms = PosTools.obtain_observation_year_month_day_fraction(starttime)
        if not proxyServer:
            if not "http" in ionex_server:      #ftp server use ftplib
                ionexf = ionex.getIONEXfile(time=date_parms,
                                              server  = ionex_server,
                                              prefix  = ionex_prefix,
                                              outpath = ionexPath)
            else:
                ionexf = ionex.get_urllib_IONEXfile(time  = date_parms,
                                              server  = ionex_server,
                                              prefix  = ionex_prefix,
                                              outpath = ionexPath)
            
        else:
            ionexf = ionex.get_urllib_IONEXfile(time  = date_parms,
                                              server  = ionex_server,
                                              prefix  = ionex_prefix,
                                              outpath = ionexPath,
                                              proxy_server = proxyServer,
                                              proxy_type   = proxyType,
                                              proxy_port   = proxyPort,
                                              proxy_user   = proxyUser,
                                              proxy_pass   = proxyPass)

        if ionexf == -1:
            if not "igsiono.uwm.edu.pl" in ionex_server:
                logging.info("cannot get IONEX data, try fast product server instead")
                if not proxyServer:
                    ionexf = ionex.get_urllib_IONEXfile(time = date_parms,
                                                     server  = "https://igsiono.uwm.edu.pl",
                                                     prefix  = "igrg",
                                                     outpath = ionexPath)
                else:
                    ionexf = ionex.get_urllib_IONEXfile(time = date_parms,
                                                     server  = "https://igsiono.uwm.edu.pl",
                                                     prefix  = "igrg",
                                                     outpath = ionexPath,
                                                  proxy_server = proxyServer,
                                                  proxy_type   = proxyType,
                                                  proxy_port   = proxyPort,
                                                  proxy_user   = proxyUser,
                                                  proxy_pass   = proxyPass)
        if ionexf == -1:
            logging.error("IONEX data not available, even not from fast product server")
            return(-1)
    
    
    if not proxyServer:
	    rmdict = getRM.getRM(MS,
                         server    = ionex_server, 
                         prefix    = ionex_prefix, 
                         ionexPath = ionexPath, 
                         timestep  = timestepRM,
                         earth_rot = earth_rot)
    else:
	    rmdict = getRM.getRM(MS,
                         server    = ionex_server, 
                         prefix    = ionex_prefix, 
                         ionexPath = ionexPath, 
                         timestep  = timestepRM,
                         earth_rot = earth_rot,
                         proxy_server = proxyServer,
                         proxy_type   = proxyType,
                         proxy_port   = proxyPort,
                         proxy_user   = proxyUser,
                         proxy_pass   = proxyPass)

    
    if not rmdict:
        if not ionex_server:
            raise ValueError("One or more IONEX files is not found on disk and download is disabled!\n"
                                 "(You can run \"bin/download_IONEX.py\" outside the pipeline if needed.)")
        else:
            raise ValueError("Couldn't get RM information from RMextract! (But I don't know why.)")
        
    logging.info('Adding rotation measure values to: ' + solset_name + ' of ' + h5parmdb)
    if all_stations:
        if type(list(station_names)[0]) != str:
            rm_vals = np.array([rmdict["RM"][stat.decode()].flatten() for stat in station_names])
        else:
            rm_vals = np.array([rmdict["RM"][stat].flatten() for stat in station_names])
    else:
        rm_vals  = np.ones((len(station_names),rmdict['RM']['st0'].shape[0]),dtype=float)
        rm_vals += rmdict['RM']['st0'].flatten()[np.newaxis]
        
    new_soltab = solset.makeSoltab(soltype='rotationmeasure', soltabName='RMextract',
                                   axesNames=['ant', 'time'], axesVals=[station_names, rmdict['times']],
                                   vals=rm_vals,
                                   weights=np.ones_like(rm_vals, dtype=np.float16))

    
    return(0)

    data.close()

    
    ########################################################################
def main():
    parser = argparse.ArgumentParser(description='Adds CommonRotationAngle to an H5parm from RMextract.')

    parser.add_argument('MSfiles', type=str, nargs='+',
                        help='MS for which the parmdb should be created.')
    parser.add_argument('h5parm', type=str,
                        help='H5parm to which the results of the CommonRotationAngle is added.')
    parser.add_argument('--server', type=str, default='ftp://ftp.aiub.unibe.ch/CODE/',
                        help='URL of the server to use. (default: ftp://ftp.aiub.unibe.ch/CODE/)')
    parser.add_argument('--prefix', type=str, default='CODG',
                        help='Prefix of the IONEX files. (default: \"CODG\")')
    parser.add_argument('--ionexpath', '--path', type=str, default='./',
                        help='Directory where to store the IONEX files. (default: \"./\")')
    parser.add_argument('--solsetName', '--solset', type=str, default='sol000',
                        help='Name of the h5parm solution set (default: sol000)')
    parser.add_argument('--single','-s', help=
                        'calculate RM only for CS002LBA (default calculate for all stations in the MS)',
                        action='store_false',dest="allStations")
    parser.add_argument('-t','--timestep', help=
                        'timestep in seconds. for values <=0 (default) the timegrid of the MS is used ',
                        dest="timestep",type=float, default=300.)
    parser.add_argument('-e','--smart_interpol', type=float, default=0, help=
                        'float parameter describing how much of Earth rotation is taken in to account \
                        in interpolation of the IONEX files. 1.0 means time interpolation assumes \
                        ionosphere rotates in opposite direction of the Earth. 0.0 (default) means \
                        no rotation applied',dest="earth_rot",)
    parser.add_argument('--proxyserver', type=str, default=None,
                        help='Name of a proxy server to use (default: None)')
    parser.add_argument('--proxyport', type=int, default=None,
                        help='Name of a proxy server port to use (default: None)')
    parser.add_argument('--proxytype', type=str, default=None,
                        help='Name of a proxy server type to use (default: None)')
    parser.add_argument('--proxyuser', type=str, default=None,
                        help='Name of a proxy server user to use (default: None)')
    parser.add_argument('--proxypass', type=str, default=None,
                        help='Password of the proxy server user to use (default: None)')


    args = parser.parse_args()

    MS = args.MSfiles
    h5parmdb = args.h5parm
    logging.info("Working on: %s %s" % (MS, h5parmdb))
    createRMh5parm(MS, h5parmdb, ionex_server=args.server, ionex_prefix=args.prefix, 
                 ionexPath=args.ionexpath, solset_name=args.solsetName,
                 all_stations=args.allStations, timestepRM=args.timestep,
                 earth_rot=args.earth_rot, proxyServer=args.proxyserver, proxyPort=args.proxyport,
                 proxyType=args.proxytype,proxyUser=args.proxyuser,proxyPass=args.proxypass)
    



import os
import sys
import warnings
from setuptools import find_packages
import numpy
from numpy.distutils.core import setup, Extension


def read(rel_path):
    """Function read() was copied from setup.py in Pip package."""
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


packages = find_packages()  # exclude=["RMextract.LOFAR_TOOLS"])
ext_modules = []

ext_modules.append(
    Extension(
        "RMextract.EMM._EMM_Model",
        sources=[
            os.path.join("RMextract", "EMM", f)
            for f in ("EMM_Model.cc", "GeomagnetismLibrary.c", "EMM_Model_wrap.cc")
        ],
        extra_compile_args=["-Wno-format-security"],
    )
)

ext_modules.append(
    Extension(
        "RMextract.pyiri._iri",
        sources=[
            os.path.join("RMextract", "pyiri", f)
            for f in (
                "iri.pyf",
                "cira.for",
                "igrf.for",
                "iridreg.for",
                "iriflip.for",
                "irifun.for",
                "irisub.for",
                "iritec.for",
                "iriget.for",
            )
        ],
        include_dirs=[numpy.get_include()],
    )
)

ext_modules.append(
    Extension(
        "RMextract.pyiriplas._iriplas",
        sources=[
            os.path.join("RMextract", "pyiriplas", f)
            for f in (
                "iriplas.pyf",
                "igrf.for",
                "irif2019.for",
                "iriplas_main.for",
                "Iris2017.for",
                "indx2017.for",
            )
        ],
        include_dirs=[numpy.get_include()],
    )
)

# For backward compatibility for those who (still) use `python setup.py` to
# install the optional LOFAR utilities.
if "--add-lofar-utils" in sys.argv:
    packages.append("RMextract.LOFAR_TOOLS")
    sys.argv.remove("--add-lofar-utils")
    warnings.warn(
        "Use of 'python setup.py install --add-lofar-utils' is deprecated. "
        "Use 'pip install RMextract[lofar-utils]' instead."
    )

setup(
    name="RMextract",
    version="0.4.3",
    url="https://github.com/lofar-astron/RMextract",
    project_urls={"Source": "https://github.com/lofar-astron/RMextract"},
    author="Maaijke Mevius",
    author_email="mevius@astron.nl",
    description="Extract TEC, vTEC, Earthmagnetic field and Rotation Measures from GPS "
                "and WMM data for radio interferometry observations",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    maintainer="Marcel Loose",
    maintainer_email="loose@astron.nl",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    ext_modules=ext_modules,
    packages=packages,
    install_requires=["numpy", "scipy", "astropy", "python-casacore"],
    extras_require={
        # Note that "lofar-utils" also depends on the python bindings to the LOFAR ParmDB.
        # Since these have never been published on PyPI, we cannot specify this dependency.
        "lofar-utils": ["losoto"]
    },
    entry_points={
        "console_scripts": [
            "createRMParmdb = RMextract.LOFAR_TOOLS.createRMParmdb:main [lofar-utils]",
            "createRMh5parm.py = RMextract.LOFAR_TOOLS.createRMh5parm:main [lofar-utils]",
            "download_IONEX.py = RMextract.LOFAR_TOOLS.download_IONEX:main [lofar-utils]",
        ]
    },
    package_data={
        "RMextract.EMM": ["*.COF"],
        # Add *.pyf files. These files are _not_ treated as source files by Numpy's setup(),
        # because, normally, they are (re)generated by f2py from the Fortran source files.
        # However, since these files needed to be tweaked a bit, they _must_ be treated as source
        # files now. So, we have to manually add them to the source distribution.
        "RMextract.pyiri": ["iri.pyf", "*.dat", "*.asc"],
        "RMextract.pyiriplas": ["iriplas.pyf", "*.dat", "*.asc", "kp*", "*.ASC"],
    },
)



from __future__ import absolute_import
import RMextract.EMM.EMM_Model as emm
from math import *
import numpy as np
from pkg_resources import resource_filename
import os

class WMM:
    def __init__(self,cof=resource_filename(__name__,'WMM.COF'),date=2010.,lon=0.,lat=0.,h=0.):
        try:
            cof=cof.replace("WMM.COF","")
            if not os.path.isfile(cof+"WMM.COF"):
                print ("initializing model failed. Did you specify correct location for coefficient files")
                return
            self.WMM_Model=emm.WMM_Model(cof+"WMM",date,lon,lat,h);
        except (RuntimeError, TypeError, NameError):
            print ("initializing mdoel failed. Did you specify correct location for coefficient files")
        self.date=date;
        self.lon=lon;
        self.lat=lat;
        self.h=h;
        self.changed=False

    def getXYZ(self):
        '''get BField, convert to (ITRF) X,Y,Z coordinates'''
        N,E,D=self.getNED()
        lat=radians(self.lat)
        lon=radians(self.lon)
        X=-1*cos(lat)*cos(lon)*D-sin(lat)*cos(lon)*N-sin(lon)*E;
        Y=-1*cos(lat)*sin(lon)*D-sin(lat)*sin(lon)*N+cos(lon)*E;
        Z=-1*sin(lat)*D+N*cos(lat)
        return X,Y,Z

    def getNED(self):
        #check if something has changed
        if self.WMM_Model.getDate() != self.date:
            self.changed=True;
            self.WMM_Model.setDate(self.date);
        if self.WMM_Model.getLon() != self.lon or self.WMM_Model.getLat() != self.lat :
            self.changed=True;
            self.WMM_Model.setLonLat(self.lon,self.lat);
        if self.WMM_Model.getHeight() != self.h:
            self.changed=True;
            self.WMM_Model.setHeight(self.h);
        if self.changed:
            self.WMM_Model.setEM()
            self.changed=False
            
        return self.WMM_Model.getX(),self.WMM_Model.getY(),self.WMM_Model.getZ()
        #North, East,Down
    
    def getProjectedField(self,lon,lat):
        '''get fieldstrength  in direction lon,lat'''
        x,y,z=self.getXYZ();
        return sin(lat)*z+cos(lat)*cos(lon)*x +cos(lat)*sin(lon)*y;

    def getProjectedFieldVector(self,lon,lat):
        '''get field  with one axis projected in direction lon,lat'''
        x,y,z=self.getXYZ();
        return [sin(lat)*z+cos(lat)*cos(lon)*x +cos(lat)*sin(lon)*y,
                -1*sin(lon)*x +cos(lon)*y,
                cos(lat)*z-sin(lat)*cos(lon)*x -sin(lat)*sin(lon)*y]

    def getProjectedFieldArray(self,lon_array,lat_array,h_array,los_dir):
        '''returns numpy array with projected BField along LOS (given in lon,lat) for ray tracing'''
        result=np.zeros((len(lon_array),3))
        
        for idx,self.lon,self.lat,self.h in zip(range(len(lon_array)),lon_array,lat_array,h_array):
            result[idx]=self.getProjectedFieldVector(los_dir[0],los_dir[1])
        return result


class EMM(WMM):
    def __init__(self,cof=resource_filename(__name__,'EMM2000.COF'),date=2010.,lon=0.,lat=0.,h=0.):
        try:
            cof=cof.replace("EMM2000.COF","")
            if not os.path.isfile(cof+"EMM2000.COF"):
                print ("initializing model",cof+"EMM2000.COF","failed. Did you specify correct location for coefficient files")
                return
            
            self.WMM_Model=emm.EMM_Model(cof+"EMM",date,lon,lat,h);
        except (RuntimeError, TypeError, NameError):
            print ("initializing mdoel failed. Did you specify correct location for coefficient files")
        self.date=date;
        self.lon=lon;
        self.lat=lat;
        self.h=h;
        self.changed=False



from . import _iri
import numpy as np
from pkg_resources import resource_filename
from scipy.interpolate import interp1d



class pyiri:
    def __init__(self,year=2016,month=5,day=12,hour=1.,lon=0.,lat=0.):
        self.lon=lon
        self.lat=lat
        self.yr=year
        self.month=month
        self.day=day
        self.hour=hour
        self.flags=np.ones(50,dtype=bool)  #to do, check which flags to set
        self.jmag=0 #geographic (0) or geomagnetic (1) coordinates
        self.datapath=resource_filename(__name__,'ig_rz.dat').replace('ig_rz.dat','')
        

    def get_profile(self,hstart,hend,hstep):
        #res=_iri.iri_sub(self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=hstart,heiend=hend,heistp=hstep)[0][0]
        res=_iri.iri_get_sub(self.datapath,self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=hstart,heiend=hend,heistp=hstep)[0][0]
        myshape=np.arange(hstart,hend+hstep,hstep).shape[0]
        return res[:myshape]

    def get_tec(self):
        #res=_iri.iri_sub(self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=hstart,heiend=hend,heistp=hstep)[0][0]
        res=_iri.iri_get_tec(self.datapath,self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=50,heiend=1500)
        return res

    def get_sub(self,hstart=399,hend=400,hstep=1):
        res=_iri.iri_get_sub(self.datapath,self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=hstart,heiend=hend,heistp=hstep)
        #res=_iri.iri_get_tec(self.datapath,self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=50,heiend=1500)
        return res

    def get_hprofile(self,h):
        #res=_iri.iri_sub(self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=hstart,heiend=hend,heistp=hstep)[0][0]
        heibeg=h[0]
        heiend=min(2000,h[-1])
        heistp=(heiend-heibeg)/999.
        res=_iri.iri_get_sub(self.datapath,self.flags,jmag=self.jmag,alati=self.lat,along=self.lon,iyyyy=self.yr,mmdd=self.month*100+self.day,dhour=self.hour,heibeg=heibeg,heiend=heiend,heistp=heistp)[0][0]
        heights = np.arange(heibeg,heiend+0.5*heistp,heistp)
        hinterpol = interp1d(heights,res,fill_value='extrapolate')
        
        return hinterpol(h)
      



import RMextract.getRM as gt
import lofar.parmdb as parmdb
import pyrap.tables as tab
import numpy as np
import argparse


def createRMParmdb(MS,parmdbname,create=True,patchname='',
                   server="ftp://cddis.gsfc.nasa.gov/gnss/products/ionex/",
                   prefix='codg',
                   ionexPath="IONEXdata/",
                   earth_rot=0,
                   timestep=900.,
                   stat_names=None):
    myParmdb=parmdb.parmdb(parmdbname,create=create)
    if create:
        myParmdb.addDefValues("Gain:0:0:Ampl",1.e-4)
        myParmdb.addDefValues("DirectionalGain:0:0:Ampl",1.)
        myParmdb.addDefValues("DirectionalGain:1:1:Ampl",1.)
        myParmdb.addDefValues("Gain:0:0:Real",1.e-4)
        myParmdb.addDefValues("Gain:1:1:Real",1.e-4)
        myParmdb.addDefValues("DirectionalGain:0:0:Real",1.)
        myParmdb.addDefValues("DirectionalGain:1:1:Real",1.)
    myMS=tab.table(MS)
    stations=tab.table(myMS.getkeyword('ANTENNA')).getcol('NAME')
    stat_pos=tab.table(myMS.getkeyword('ANTENNA')).getcol('POSITION')
    if not stat_names==None:
        if not(stat_names=='all'):
            stat_pos=[stat_pos[stations.index(name)] for name in stat_names] 
        else:
            stat_names=stations[:]
    else:
        stat_names=stations[2:3]
        stat_pos=stat_pos[2:3]
    result=gt.getRM(MS=MS,server=server,ionexPath=ionexPath,prefix=prefix,earth_rot=earth_rot,timestep=timestep,stat_names=stat_names,stat_positions=stat_pos)
    RM=result['RM']    
    for st in stations:
        #print "storing station",st,(st in RM.keys())
        if not (st in RM.keys()):
            stname=RM.keys()[0]
        else:
            stname=st
        RM[stname]=RM[stname].reshape(RM[stname].shape[:1]+(1,))
        myValue=myParmdb.makeValue(values=RM[stname], sfreq=1e10, efreq=2e10, stime=result['times'], etime=np.ones(result['times'].shape,dtype=float)*result['timestep'], asStartEnd=False)
        if patchname:
            valuename = "RotationMeasure:%s:%s"%(st,patchname)
        else:
            valuename = "RotationMeasure:%s"%(st)
        myParmdb.deleteValues(valuename)            
        myParmdb.addValues(valuename,myValue)

def main():
    descriptiontext = "Create a parmDB with the Ionospheric RM predicted from data in IONEX files.\n" + \
                      "Default is to create a parmDB that is compatible with the \"rotationmeasure\" " + \
                      "correction in NDPPP ApplyCal.\n"
    parser = argparse.ArgumentParser(description=descriptiontext)
   
    parser.add_argument('MS',help='Measurement-Set for which the parmDB is to be created.')
    parser.add_argument('-o','--out', help='name of the parmdb',dest="parmdbName",default='RMParmdb');
    parser.add_argument('-p','--patch',dest="patch",default="",type=str,
                        help='If given: create parameter-names that include a patch-name as needed for BBS. '
                        'The default is to create parameter-names without a patch-name as they are needed for '
                        'the "rotationmeasure" correction in NDPPP ApplyCal.')
    #parser.add_argument('-s','--sky', help='name of the skymodel, if no source/patch name is given first will be selected. If you do not set the skymodel, the phasecenter of the MS will be assumed as your direction. If option --use_phase_center is used the direction in the sky model are ignored. However, it could still be useful to automatically select a patch name, which is needed for direction dependent parmareters.',dest="sky")
    #parser.add_argument('-f','--use_phase_center', help='force using of phase center of MS, even if skymodel is supplied',action='store_true',dest="usePhaseCenter");
    parser.add_argument('--IONprefix', help='prefix of IONEX files, either CODG or ROBR',dest="prefix",default='codg');
    parser.add_argument('--IONserver', help='server of IONEX files',dest="server",default='ftp://cddis.gsfc.nasa.gov/gnss/products/ionex/')
    parser.add_argument('--IONpath', help='location of IONEX files',dest="ionexPath",default='./')
    parser.add_argument('--all','-a', help='calculate RM per station (default calculates only for CS002LBA)',action='store_true',dest="allStations")
    parser.add_argument('-t','--timestep', help='timestep in seconds. for values <=0 (default) the timegrid of the MS is used ',dest="timestep",type=float,default=900.)
    parser.add_argument('-e','--smart_interpol', help='float parameter describing how much of earth rotation is taken in to account in interpolation of the IONEX files. 1.0 means time interpolation assumes ionosphere rotates in opposite direction of the Earth. 0.0 (default) means no rotation applied',dest="earth_rot",type=float,default=0)
    parser.add_argument('-r','--overwrite',help='overwrite existing parmdb',action='store_true',dest="overwrite")

    args = parser.parse_args()
        
    stat_names=None
    if args.allStations:
        stat_names='all'
    if not args.overwrite and not os.path.isdir(args.parmdbName):
        args.overwrite=True
        
    createRMParmdb(args.MS,
                   parmdbname=args.parmdbName,
                   create=args.overwrite,
                   patchname=args.patch,
                   server=args.server,
                   prefix=args.prefix,
                   ionexPath=args.ionexPath,
                   earth_rot=args.earth_rot,
                   timestep=args.timestep,
                   stat_names=stat_names)
